import json
import os

from datenguidepy.query_execution import ExecutionResults, QueryExecutioner


def construct_execution_results(path):
    directory, file = os.path.split(path)
    meta_file = file.split(".", 1)[0] + "_meta.json"
    meta_path = os.path.join(directory, meta_file)
    with open(path, "r", encoding="utf-8") as file:
        res = json.load(file)
    with open(meta_path, "r", encoding="utf-8") as file:
        meta = json.load(file)
    return [ExecutionResults(r, meta) for r in res]


def save_result_files_from_query(query, name, result_path):
    """
        Helper function that for the construction of files
        to be loaded by `construct_execution_results`
    """
    qe = QueryExecutioner()
    execution_results = qe.run_query(query)

    query_json_file = os.path.join(result_path, f"{name}_query.json")
    result_json_file = os.path.join(result_path, f"{name}.json")
    meta_json_file = os.path.join(result_path, f"{name}_meta.json")

    with open(query_json_file, "w") as file:
        json.dump(query.get_graphql_query(), file)

    with open(result_json_file, "w") as file:
        json.dump([res.query_results for res in execution_results], file)

    with open(meta_json_file, "w") as file:
        json.dump(execution_results[0].meta_data, file)



from typing import Dict

from abc import abstractmethod, ABC
from pandas import DataFrame
import json
import os


class TranslationProvider(ABC):
    @abstractmethod
    def translate_data_frame_from_german(
        self, data_frame: DataFrame, target_language: str
    ):
        pass

    @abstractmethod
    def translate_from_german(self, source_text: str, target_language: str):
        pass

    @abstractmethod
    def is_valid_language_code(self, target_language: str):
        pass

    @abstractmethod
    def get_valid_language_codes(self):
        pass


class SchemaTranslationProvider(TranslationProvider):
    def __init__(self):
        self.translated_schema: Dict = self.get_translated_schema_from_file()

    @staticmethod
    def get_translated_schema_from_file():
        curr_dir = os.path.dirname(os.path.abspath(__file__))
        data_dir = os.path.join(curr_dir, "package_data")
        translation_json_path = os.path.join(data_dir, "translated_schema.json")
        with open(translation_json_path) as translation_file:
            translation_schema = json.loads(translation_file.read())
        return translation_schema

    def translate_data_frame_from_german(
        self, data_frame: DataFrame, target_language: str
    ):
        data_frame.replace(self.translated_schema[target_language], inplace=True)

    def translate_from_german(self, source_text: str, target_language: str):
        language_specific_schema = self.translated_schema[target_language]
        if source_text in language_specific_schema:
            return language_specific_schema[source_text]
        else:
            return source_text

    def is_valid_language_code(self, target_language: str):
        return target_language in self.get_valid_language_codes()

    def get_valid_language_codes(self):
        return list(self.translated_schema.keys())


DEFAULT_TRANSLATION_PROVIDER = SchemaTranslationProvider()



from typing import Optional, Union, List, Dict, Any, Tuple, Set
from pandas import DataFrame
from datenguidepy.query_execution import (
    QueryExecutioner,
    GraphQlSchemaMetaDataProvider,
    StatisticsMetaDataProvider,
    DEFAULT_STATISTICS_META_DATA_PROVIDER,
    TypeMetaData,
    QueryResultsMeta,
)
from datenguidepy.output_transformer import QueryOutputTransformer


class Field:
    """A field of a query that specifies a statistic
    (or another information, e.g. source) to query.
    The name of the field (mostly statistic), the filters (specified with args)
    and the desired output information (fields)
    are specified.

    :param name: Name of Field (statistic)
    :type name: str
    :param fields: desired output fields (e.g. year or NAT), defaults to []
    :type fields: list, optional
    :param args: Filters for the desired field (e.g. {'year': 2017}).
    If "ALL" is passed as a value, then results are returned for all possible subgroups.
    (e.g. for gender 'GES': 'ALL' three data entries are returned - for
    male, female and summed for both.
    if the filter is not set, then only the summed result is returned.
    Except for year: this is by default returned for each year), defaults to {}
    :type args: dict, optional
    :param parent_field: The field this field is attached to, defaults to None
    :type parent_field: Field, optional
    :param default_fields: Wether default fields should be attached or not,
    defaults to True
    :type default_fields: bool, optional
    :param return_type: The graphQL return type of this field, defaults to None
    :type return_type: str, optional
    """

    def __init__(
        self,
        name: str,
        fields: List[Union[str, "Field"]] = [],
        args: Dict[str, Any] = {},
        parent_field: "Field" = None,
        default_fields: bool = True,
        return_type: str = None,
        stat_meta_data_provider: StatisticsMetaDataProvider = None,
    ):
        self.name = name
        self.parent_field = parent_field
        self.default_fields = default_fields
        # TODO: use name as default?
        self.return_type = return_type if return_type else name

        self.fields: Dict[str, "Field"] = {}
        if stat_meta_data_provider is None:
            self._stat_meta_data_provider: StatisticsMetaDataProvider = (
                DEFAULT_STATISTICS_META_DATA_PROVIDER
            )
        else:
            self._stat_meta_data_provider = stat_meta_data_provider

        self._graphql_schema_meta_data_provider = GraphQlSchemaMetaDataProvider()

        for field in fields:
            self.add_field(field, default_fields=default_fields)

        if default_fields and self._stat_meta_data_provider.is_statistic(self.name):

            self.fields["year"] = Field(
                "year", return_type=self._get_return_type("year")
            )
            self.fields["value"] = Field(
                "value", return_type=self._get_return_type("value")
            )
            self.fields["source"] = Field(
                "source",
                fields=["title_de", "valid_from", "periodicity", "name", "url"],
                return_type=self._get_return_type("source"),
            )

        self.args = args

    def _get_return_type(self, fieldname):
        return (
            self._graphql_schema_meta_data_provider.get_type_info(self.return_type)
            .fields[fieldname]
            .get_return_type()
        )

    def add_field(
        self, field: Union[str, "Field"], default_fields: bool = None
    ) -> "Field":
        """Add a subfield to the field.

        :raises TypeError: If the added field is neither of type String nor Field.
        :return: the added field
        :rtype: Field
        """
        if default_fields is None:
            default_fields = self.default_fields

        if isinstance(field, str):
            self.fields[field] = Field(
                name=field,
                parent_field=self,
                fields=[],
                return_type=self._get_return_type(field),
                default_fields=default_fields,
                stat_meta_data_provider=self._stat_meta_data_provider,
            )
            return self.fields[field]
        else:
            field._set_return_type(self)
            field._set_parent_field(self)
            field._stat_meta_data_provider = self._stat_meta_data_provider
            self.fields[field.name] = field
            return self.fields[field.name]

    def drop_field(self, field: str) -> "Field":
        """Drop an attached subfield of the field.

        :param field: The name of the field to be droped.
        :type field: str
        :return: The field without the subfield.
        :rtype: Field
        """
        if isinstance(field, str):
            self.fields.pop(field, None)
        else:
            self.fields.pop(field.name, None)
        return self

    def add_args(self, args: dict):
        """Add arguments to the field.
        :param args: Arguments to be added.
        :type args: dict
        """
        if self.args:
            self.args.update(args)
        else:
            self.args = args

    def _set_return_type(self, parentfield):
        self.return_type = parentfield._get_return_type(self.name)

    def _set_parent_field(self, parent_field):
        self.parent_field = parent_field

    def _get_fields_to_query(self, field: "Field", region_id: str = None) -> str:
        substring = field.name + " "

        if field.args:
            # make copy, so original field id is not overwritten
            this_query_args = field.args.copy()

            if ("id" in field.args) & (region_id is not None):
                # set region id to given single id to not use list
                this_query_args["id"] = region_id

            filters = []
            for key, value in this_query_args.items():
                if value == "ALL":
                    filters.append("filter:{ " + key + ": { nin: []}}")
                else:
                    # delete quotation marks for query arguments
                    filters.append(key + ": " + str(value).replace("'", ""))
            substring += "(" + ", ".join(filters) + ")"

        if field.fields:
            substring += "{"
            for field_item in field.fields.values():
                substring += field._get_fields_to_query(field_item)
            substring += "}"
        return substring

    def get_fields(self) -> List[str]:
        """Get all fields that are attached to this
        field or subfields of this field.

        :return: a list of all fields
        :rtype: List[str]
        """

        field_list = [self.name]
        for value in self.fields.values():
            field_list.extend(Field._get_fields_recursion(value))
        return field_list

    def _get_fields_with_types(self) -> List[Tuple[str, str]]:
        """ Gets all the fields and attached to
            this field including all subfields. Additionally
            returns the return type for each field. This will
            allow internal functions to easily reques meta
            data for specific fields.

            :return: a list of tuples with field names and
            their types
            :rtype: List[Tuple[str,str]]
        """
        fields_with_types = [(self.name, self.return_type)]
        for field in self.fields.values():
            fields_with_types.extend(field._get_fields_with_types())
        return fields_with_types

    def get_info(self) -> None:
        """Prints summarized information on a field's meta data.

        :return: None
        :rtype: None
        """

        kind: str = _bold_font("kind:") + "\n"
        meta = self._graphql_schema_meta_data_provider.get_type_info(self.name)
        if meta is not None:
            kind += meta.kind
        else:
            kind += "None"
        description = _bold_font("description:") + "\n" + str(self.description())
        arguments = _bold_font("arguments:") + "\n" + str(self.arguments_info())
        fields = _bold_font("fields:") + "\n" + str(self.fields_info())
        enum_values = _bold_font("enum values:") + "\n" + str(self.enum_info())
        print("\n\n".join([kind, description, arguments, fields, enum_values]))

    @staticmethod
    def _no_none_values(base_function, dict_content, sub_key) -> Optional[str]:
        if dict_content is None:
            return None
        value = getattr(dict_content, sub_key)
        if value is None:
            return None
        return base_function(value)

    def _arguments_info_formatter(self, meta_fields) -> Optional[str]:
        UNDERLINE = "\033[4m"
        NORMAL = "\033[0m"
        args = meta_fields[self.name].get_arguments()
        arg_list = []

        for key, value in args.items():

            single_arg_string = UNDERLINE + key + NORMAL + ": " + str(value[0])

            # get type
            if value[0] == "LIST":
                single_arg_string += (
                    " of type " + str(value[2]) + "(" + str(value[3]) + ")"
                )
            else:
                single_arg_string += "(" + str(value[1]) + ")"

            # get enum values
            if "ENUM" in value:
                args_field = Field(name=key, parent_field=self, return_type=value[3])
                enum_values = args_field.enum_info()
                single_arg_string += "\nenum values:\n" + str(enum_values)
            arg_list.append(single_arg_string)

        return "\n\n".join(arg_list)

    def arguments_info(self) -> Optional[str]:
        """Get information on possible arguments for field. The name of the argument is
        followed by the kind and name of the input type for the argument in brackets.
        If the argument is a list, the kind and name of the list elements are
        included in the brackets as well.

        :return: Possible arguments for the field as string and their input types.
        :rtype: Optional[str]
        """

        parent = self.parent_field
        if parent is not None:
            meta = self._graphql_schema_meta_data_provider.get_type_info(
                parent.return_type
            )
            return Field._no_none_values(self._arguments_info_formatter, meta, "fields")
        else:
            return None

    def _fields_info_formatter(self, meta_fields) -> Optional[str]:
        args_info = []
        for meta_field in meta_fields:
            args_info.append(meta_field + ": " + meta_fields[meta_field]["description"])
        return "\n".join(args_info)

    def fields_info(self) -> Optional[str]:
        """Get information on possible fields for field.

        :return: Possible fields for the field as string
        :rtype: Optional[str]
        """

        meta = self._graphql_schema_meta_data_provider.get_type_info(self.name)
        return Field._no_none_values(self._fields_info_formatter, meta, "fields")

    def _enum_info_formatter(self, enum_meta) -> Optional[str]:
        enum_list = []
        for key, value in enum_meta.items():
            enum_list.append(key + ": " + value)
        return "\n".join(enum_list)

    def enum_info(self) -> Optional[str]:
        """Get information on possible enum vaules for field.

        :return:  Possible enum values for the field as string.
        :rtype: Optional[str]
        """

        meta = self._graphql_schema_meta_data_provider.get_type_info(self.return_type)
        return Field._no_none_values(self._enum_info_formatter, meta, "enum_values")

    def description(self) -> Optional[str]:
        """Get description of field.

        :return: Description of the field as string.
        :rtype: Optional[str]
        """
        return self._stat_meta_data_provider.get_stat_descriptions()[self.name][0]

    @staticmethod
    def _get_fields_recursion(field: "Field") -> List[str]:
        field_list = []
        field_list.append(field.name)
        if field.fields:
            for value in field.fields.values():
                field_list.extend(Field._get_fields_recursion(value))
        return field_list


def _bold_font(text: str) -> str:
    BOLD = "\033[1m"
    NORMAL = "\033[0m"
    return BOLD + text + NORMAL


class Query:
    """A query to get information via the datenguide API for regionalstatistik.
    The query contains all fields and arguments.

    :param start_field: The top node field; either allRegions or Region.
    :type start_field: Field
    :param region_field: A field of type 'Region' that is needed
    if start_field is allRegions, defaults to None
    :type region_field: Field, optional
    :param default_fields: Wether default fields shall
            be attached to the fields., defaults to True
    :type default_fields: bool, optional
    :raises RuntimeError: [description]
    """

    # static variables specific to datenguide graphQL API
    _return_type_region: str = "Region"
    _return_type_allreg: str = "RegionsResult"
    _return_type_regions: str = "Region"

    def __init__(
        self,
        start_field: Field,
        region_field: Field = None,
        default_fields: bool = True,
        stat_meta_data_provider: StatisticsMetaDataProvider = None,
    ):
        self.start_field = start_field
        self.region_field = region_field
        self.result_meta_data: Optional[QueryResultsMeta] = None
        if stat_meta_data_provider is None:
            self._stat_meta_data_provider: StatisticsMetaDataProvider = (
                DEFAULT_STATISTICS_META_DATA_PROVIDER
            )
        else:
            self._stat_meta_data_provider = stat_meta_data_provider

        self._graphql_schema_meta_data_provider = GraphQlSchemaMetaDataProvider()

    @classmethod
    def region(
        cls,
        region: Union[str, List[str]],
        fields: List[Union[str, "Field"]] = [],
        default_fields: bool = True,
        stat_meta_data_provider=None,
    ) -> "Query":
        """Factory method to instantiate a Query with a single region through
            its region id.

        :param region: The region id(s) the statistics shall return
        :type region: Union[str, List[str]]
        :param fields: all fields that shall be
                returned from the query for that region.
                Can either be simple fields (e.g. name)
                or fields with nested fields.
        :type fields: list
                or fields with nested fields.
        :param default_fields: Wether default fields shall
        :type default_fields: bool

        :raises RuntimeError: [description]

        :return: A query object with region as start Field.
        :rtype: Query
        """

        if default_fields:
            defaults: List[Union[str, "Field"]] = ["id", "name"]
            fields = defaults + fields

        # add quotation marks around id for correct query
        if isinstance(region, list):
            region_arg: Union[str, List[str]] = [('"' + x + '"') for x in region]
        else:
            region_arg = ['"' + region + '"']

        return cls(
            start_field=Field(
                "region",
                fields,
                args={"id": region_arg},
                return_type=Query._return_type_region,
                default_fields=default_fields,
                stat_meta_data_provider=stat_meta_data_provider,
            ),
            stat_meta_data_provider=stat_meta_data_provider,
        )

    @classmethod
    def all_regions(
        cls,
        fields: List[Union[str, "Field"]] = [],
        parent: str = None,
        nuts: int = None,
        lau: int = None,
        default_fields: bool = True,
        stat_meta_data_provider=None,
    ) -> "Query":
        """Factory method to instantiate a Query with allRegions start field.
        A parent id, nuts or lau can be further specified for the query.

        :param fields: all fields that shall be returned
            for that region. Can either be simple fields (e.g. name)
            or fields with nested fields.
        :param parent: The region id of the parent region
            the statistics shall be queried for.
            (E.g. the id for a state where all sub regions within the
            state shall be queried for.)
        :type parent:
        :param nuts: The administration level: 1 – Bundesländer
            2 – Regierungsbezirke / statistische Regionen
            3 – Kreise / kreisfreie Städte.
            Default None returns results for all levels.
        :type nuts: int, optional
        :param lau: The administration level: 1 - Verwaltungsgemeinschaften
            2 - Gemeinden.
            Default returns results for all levels.
        :type lau: int, optional
        :type fields: list
        :param default_fields: Wether default fields shall
            be attached to the fields.
        :type default_fields: bool


        :return:  A query object with allRegions as start Field.
        :rtype: Query
        """

        # add page and itemsPerPage as arguments for QueryExecutioner
        args = {"page": "$page", "itemsPerPage": "$itemsPerPage"}

        region_args = {}
        if parent:
            region_args["parent"] = '"' + parent + '"'
        if nuts:
            region_args["nuts"] = str(nuts)
        if lau:
            region_args["lau"] = str(lau)

        if default_fields:
            defaults: List[Union[str, "Field"]] = ["id", "name"]
            fields = defaults + fields

        regions = Field(
            "regions",
            fields=fields,
            args=region_args,
            return_type=Query._return_type_regions,
            default_fields=default_fields,
            stat_meta_data_provider=stat_meta_data_provider,
        )

        # add fields page, itemsperPage and total for QueryExecutioner
        return cls(
            start_field=Field(
                "allRegions",
                fields=[regions, "page", "itemsPerPage", "total"],
                args=args,
                return_type=Query._return_type_allreg,
                default_fields=default_fields,
                stat_meta_data_provider=stat_meta_data_provider,
            ),
            region_field=regions,
            stat_meta_data_provider=stat_meta_data_provider,
        )

    def add_field(
        self, field: Union[str, Field], default_fields: bool = None
    ) -> "Field":
        """Add a field to the query.

        :param field: Field to be added
        :type field: Union[str, Field]
        :param default_fields: Wether default fields
        should be attached or not, defaults to None
        :type default_fields: bool, optional
        :raises RuntimeError: If the allRegions Query has
            no regions field a subfield can be attached to.
        :return: The added field.
        :rtype: Field
        """

        if default_fields is None:
            default_fields = self.start_field.default_fields

        if self.start_field.name == "allRegions":
            if self.region_field is not None:
                return self.region_field.add_field(field, default_fields=default_fields)
            else:
                raise RuntimeError(
                    "All Regions Query initialized without regions field."
                )
        else:
            return self.start_field.add_field(field, default_fields=default_fields)

    def drop_field(self, field: str) -> "Query":
        """Drop an attached field of the query.

        :param field: The name of the field to be droped
        :type field: str
        :raises RuntimeError: Raises Error if Query is
        initialized without regions field.
        :return: the query without the dropped field
        :rtype: Query
        """

        if self.start_field.name == "allRegions":
            if self.region_field is not None:
                self.region_field.drop_field(field)
                return self
            else:
                raise RuntimeError(
                    "All Regions Query initialized without regions field."
                )
        else:
            self.start_field.drop_field(field)
            return self

    def get_graphql_query(self) -> List[str]:
        """Formats the Query into a String that can be queried from the Datenguide API.

        :return: the Query formatted for the GraphQL API as a List of query strings
        :rtype: List[str]
        """
        if self.start_field.name == "allRegions":
            query_prefix = "query ($page : Int, $itemsPerPage : Int) "
        else:
            query_prefix = ""

        # for region with multiple region IDs return a list of queries
        if (self.start_field.name == "region") and isinstance(
            self.start_field.args.get("id", ""), list
        ):
            query_list: List[str] = []
            for region_id in self.start_field.args["id"]:
                query_list += [
                    (
                        "{"
                        + self.start_field._get_fields_to_query(
                            self.start_field, region_id
                        )
                        + "}"
                    )
                ]
            return query_list
        else:
            return [
                query_prefix
                + "{"
                + self.start_field._get_fields_to_query(self.start_field)
                + "}"
            ]

    def get_fields(self) -> List[str]:
        """Get all fields of a query.

        :return: a list field names
        :rtype: List[str]
        """
        return self.start_field.get_fields()

    def _get_fields_with_types(self) -> List[Tuple[str, str]]:
        """Get all fields of a query including their
        return type.

        :return: a list of tuples with
            field names and their types.
        :rtype: List[Tuple[str, str]]
        """
        return self.start_field._get_fields_with_types()

    def results(
        self,
        verbose_statistics: bool = False,
        verbose_enums: bool = False,
        add_units: bool = False,
        remove_duplicates: bool = True,
    ) -> DataFrame:
        """Runs the query and returns a Pandas DataFrame with the results.
           It also fills the instance variable result_meta_data with meta
           data specific to the query instance.

        :param verbose_statistics: Toggles whether statistic column names
            displayed with their short description in the result data frame
        :param verbose_enums: Toggles whether enum values are displayed
            with their short description in the result data frame
        :param add_units: Adds units available in the metadata to the
            result dataframe. Care should be taken, because not every
            statistic specifies these corretly. When in doubt one should
            refer to the statistic description.
        :param remove_duplicates: Removes duplicates from query results, i.e. if the
            exact same number has been reported for the same statistic, year, region
            etc. from the same source it gets removed. Such duplications are sometimes
            caused on the API side and this is convenience functionality to remove them.
            The removal happens before potentially joining several different statistics.
            Unless diagnosing the API the default (True) is generally in the users
            interest.

        :raises RuntimeError: If the query fails raise RuntimeError.
        :return: A DataFrame with the queried data.
        :rtype: DataFrame
        """
        if not self._contains_statistic_field():
            raise Exception(
                "No statistic field is defined in query, please add statistic field "
                "via method add_field."
            )

        result = QueryExecutioner(
            statistics_meta_data_provider=self._stat_meta_data_provider
        ).run_query(self)
        if result:
            # It is currently assumed that all graphql queries
            # that are generated internally for the Query instance
            # at hand yield the same meta data.
            if self._query_result_contains_undefined_region(result):
                raise ValueError("Queried region is invalid.")
            self.result_meta_data = result[0].meta_data
            return QueryOutputTransformer(result).transform(
                verbose_statistic_names=verbose_statistics,
                verbose_enum_values=verbose_enums,
                add_units=add_units,
                remove_duplicates=remove_duplicates,
            )
        else:
            raise RuntimeError("No results could be returned for this Query.")

    def _query_result_contains_undefined_region(self, result):
        return (
            len(
                list(filter(lambda res: res.contains_undefined_region_result(), result))
            )
            > 0
        )

    def _contains_statistic_field(self) -> bool:
        fields = self._get_all_field_names()
        contains_statistic = any(
            [self._stat_meta_data_provider.is_statistic(field) for field in fields]
        )
        return contains_statistic

    def _get_all_field_names(self) -> Set[str]:
        start_field_subfields = (
            set() if self.start_field is None else set(self.start_field.fields.keys())
        )
        region_field_subfields = (
            set() if self.region_field is None else set(self.region_field.fields.keys())
        )
        return start_field_subfields.union(region_field_subfields)

    def meta_data(self) -> QueryResultsMeta:
        """Runs the query and returns a Dict with the meta data of the queries results.

        :raises RuntimeError: If the Query did not return any results.
        E.g. if the Query was ill-formed.
        :return: A Dict with the queried meta data.
            If the query fails raise RuntimeError.
        :rtype: Union[Dict[str, Any], List[Dict[str, Any]]]
        """

        result = QueryExecutioner(
            statistics_meta_data_provider=self._stat_meta_data_provider
        ).run_query(self)
        if result:
            # TODO: correct indexing?
            return result[0].meta_data
        else:
            raise RuntimeError("No results could be returned for this Query.")

    def get_info(self, field: str = None) -> Optional[TypeMetaData]:
        """Get information on a specific field.
        If field is not specified return meta data for
        all statistics that can be queried.

        :param field: the field to get information on. If None,
            then information on all possible fields of a query are
            returned, defaults to None
        :type field: str, optional
        :return: Response from QueryExecutioner on meta data info
        :rtype: Optional[TypeMetaData]
        """
        if field:
            return self._graphql_schema_meta_data_provider.get_type_info(field)
        else:
            return self._graphql_schema_meta_data_provider.get_type_info(
                Query._return_type_region
            )



import pytest
import pandas as pd
import sys
import io
import re
from datenguidepy.query_execution import (
    QueryExecutioner,
    StatisticsGraphQlMetaDataProvider,
)
from datenguidepy.query_builder import Query, Field
from datenguidepy.query_helper import (
    federal_states,
    get_statistics,
    get_regions,
    download_all_regions,
)
from datenguidepy.output_transformer import QueryOutputTransformer


@pytest.fixture
def query():
    field = Field(name="BEVMK3", fields=["value", "year"])
    query = Query.region(region="05911", fields=["id", "name", field])
    return query


@pytest.fixture
def query_multi_regions():
    field = Field(name="BEVMK3", fields=["value", "year"])
    query = Query.region(region=["01", "02"], fields=["id", "name", field])
    return query


@pytest.fixture
def query_all_regions():
    field = Field(name="BEVMK3", fields=["value", "year"])
    query = Query.all_regions(nuts=1, fields=["id", "name", field])
    return query


def test_query_executioner_workflow(query):
    """Functional test for the query executioner"""

    # Ira (W. Cotton, probably the first person to use the term API) want to
    # try the query execution part of the datenguide GraphQL wrapper library.
    # He already worked through Query builder objects ready for execution.
    # He understands that he first has to create an executioner object that
    # it uses the right by default, so that he does not have to supply any
    # parameters.

    q_exec = QueryExecutioner()

    # After creating the object Ira is actually a little sceptical whether
    # the endpoint will correct so he extracts the endpoint and compares it
    # with his expectations.

    assert (
        q_exec.endpoint == "https://api-next.datengui.de/graphql"
    ), "Default endpoint is wrong"

    # Being satisfied that everything is setup with the correct endpoint Ira
    # now wants to execute one of his queries to see that he gets some return
    # values.

    res_query1 = query.results()

    assert res_query1 is not None, "query did not return results"

    # He wants to have a closer look at the raw return query results and
    # remembers that they are sorted in the results field and he has a look.

    assert (
        type(res_query1) is pd.DataFrame
    ), "query results are not a python json representation"

    stats = query.get_info()

    # Ira remembers that he read about the executioners functionality to
    # return metadata along with the query results. So he wants to check
    # whether this metadata is actually present. And that it only contains
    # meta data related to his query

    meta_query1 = query.meta_data()
    meta_query1_alternative = query.result_meta_data

    # In particular Ira would like to have a more human readable description
    # of the statistic he asked for.

    assert "BEVMK3" in meta_query1["statistics"], "statistic absend"
    assert (
        meta_query1["statistics"]["BEVMK3"] != "NO DESCRIPTION FOUND"
    ), "descrption was not obtained"
    assert "BEVMK3" in meta_query1["units"], "units absend"
    assert meta_query1["units"]["BEVMK3"] == "Anzahl", "correct unit was not obtained"
    assert (
        meta_query1 == meta_query1_alternative
    ), "meta_data_query_alternatives_should be equal"

    # Although he is satisfied with having access to the meta information
    # already he would like to try the functionality where this information
    # is used directly to get more verbose query results.

    res_query1_verbose_cols = query.results(verbose_statistics=True)
    print(res_query1_verbose_cols.columns)
    assert (
        # "Von der Scheidung betroffene Kinder (BEVMK3)" in res_query1_verbose_cols
        "Von der Scheidung betroffene Kinder (BEVMK3)"
        in res_query1_verbose_cols
    ), "verbose statistic name is not present"

    # Being satisfied with the results he obtained for his simple query
    # he actually wants to try a larger one across several regions. He heard
    # that this might be an issue for the server in general, but that the
    # executioner takes care of addressing this issue by itself.

    # Since this is a lot of information Ira would particularly
    # like to drill down on the arguments that are allowed for his
    # favorite statistic BEVMK3

    stat_args = stats.fields["BEVMK3"].get_arguments()
    assert len(stat_args) > 0
    assert "statistics" in stat_args

    # Although this is already really helpful Ira notices that
    # one of the arguments is an ENUM and he would like to know
    # the possible values that he can use for it.

    enum_vals = query.get_info("BEVMK3Statistics").enum_values
    assert type(enum_vals) == dict, "Enum values should be dict"
    assert len(enum_vals) > 0, "Enums should have values"

    # Ira wants to add another statistic to his query.
    statistic1 = query.add_field("BEV001")
    statistic1.add_field("year")
    statistic1.add_args({"year": 2017})

    assert type(statistic1) == Field, "statistic is not a Field"

    # Then he wants to get metainfo on the field.

    stringio = io.StringIO()
    sys.stdout = stringio

    statistic1.get_info()
    stats_info = re.sub(r"\n", "", stringio.getvalue())
    assert "OBJECT" in stats_info, "BEV001 should be an object"


# does not require internet in this for and does not use the graphql API
# this could be changed by explicitly using the graphql statistics meta data provider
def test_query_helper():
    # Ira is happy with the functionality so far but is worried a bit
    # about the difficuilty of finding the right technichal ids for regions
    # and statistics.

    # He realizes that there is helper functionality to identify the
    # federal states quickly in a human readable way and wants to try
    # it for Berlin
    assert federal_states.Berlin == "11"

    # That already worked nicely but in general there are many regions.
    # Ira would like to easily search through all of them and realizes
    # that he can obtain a DataFrame for this.
    reg_locally_stored = get_regions()
    assert isinstance(reg_locally_stored, pd.DataFrame)

    # Being satisfied with the regions, Ira now wants to have a
    # look at an equivalent overview of statistics.

    statistics = get_statistics()
    assert isinstance(statistics, pd.DataFrame)

    # Although this might already be sufficient for finding
    # interesting statistics, Ira read that there is already
    # some basic build in search functionality which
    # he wants to try.

    filtered_statistics = get_statistics("scheidung")
    assert isinstance(filtered_statistics, pd.DataFrame)
    assert filtered_statistics.shape[0] < 50


@pytest.mark.skip(reason="currently takes very long")
def downloade_all_regions():
    # Ira reads in the help that the regions dataframe
    # is a stored in a local file and
    # not obtained live from datenguide.
    # He knows that region definitions and ids don't change very
    # often, but he would like the ability to obtain the most up to date
    # regions anyways. He therefore tries the function download_all_regions
    # that is designed for this purpouse

    reg = download_all_regions()
    assert isinstance(reg, pd.DataFrame)
    assert list(reg.columns) == ["name", "level", "parent"]
    assert reg.index.name == "id"
    assert reg.shape[0] > 10000


def test_build_execute_transform_integration(query):
    """
    Smoke test covering region query.
    """

    q_exec = QueryExecutioner()

    res = q_exec.run_query(query)

    output_transf = QueryOutputTransformer(res)
    output_transf.transform()


def test_build_execute_transform_integration_multi_region(query_multi_regions):
    """
    Smoke test covering multiple regions in
    region query.
    """

    q_exec = QueryExecutioner()

    res = q_exec.run_query(query_multi_regions)

    output_transf = QueryOutputTransformer(res)
    output_transf.transform()


def test_build_execute_transform_integration_all_regions(query_all_regions):
    """
    Smoke test covering all_regions
    """
    q_exec = QueryExecutioner()

    res = q_exec.run_query(query_all_regions)

    output_transf = QueryOutputTransformer(res)
    output_transf.transform()


def test_get_query_specific_stat_meta():
    field_type_list = [
        ("id", "String"),
        ("name", "String"),
        ("WAHL09", "WAHL09"),
        ("PART04", "PART04"),
        ("year", "Int"),
        ("value", "Float"),
    ]
    query_stat_meta = StatisticsGraphQlMetaDataProvider().get_query_stat_meta(
        field_type_list
    )
    # "Gültige Zweitstimmen" uninformative due to API changes
    expected_stat_meta = {"WAHL09": "WAHL09"}
    assert query_stat_meta == expected_stat_meta


def test_get_query_specific_enum_meta():
    field_type_list = [
        ("id", "String"),
        ("name", "String"),
        ("WAHL09", "WAHL09"),
        ("PART04", "PART04"),
        ("year", "Int"),
        ("value", "Float"),
    ]
    query_stat_meta = StatisticsGraphQlMetaDataProvider().get_query_enum_meta(
        field_type_list
    )
    expected_stat_meta = {
        "PART04": dict(
            [
                ("AFD", "AfD"),
                ("B90_GRUENE", "GRÜNE"),
                ("CDU", "CDU/CSU"),
                ("DIELINKE", "DIE LINKE"),
                ("FDP", "FDP"),
                ("SONSTIGE", "Sonstige Parteien"),
                ("SPD", "SPD"),
                ("GESAMT", "Gesamt"),
            ]
        )
    }
    assert query_stat_meta == expected_stat_meta


def test_get_query_specific_unit_meta():

    field_type_list = [
        ("id", "String"),
        ("name", "String"),
        ("WAHL09", "WAHL09"),
        ("PART04", "PART04"),
        ("year", "Int"),
        ("value", "Float"),
    ]
    query_stat_meta = StatisticsGraphQlMetaDataProvider().get_query_unit_meta(
        field_type_list
    )

    expected_stat_meta = {
        "WAHL09": "StatisticsGraphQlMetaDataProvider does not provide unit information."
    }
    assert query_stat_meta == expected_stat_meta



from datenguidepy.query_builder import Query
from datenguidepy.query_execution import (
    QueryExecutioner,
    ExecutionResults,
    DEFAULT_STATISTICS_META_DATA_PROVIDER,
)
from datenguidepy.translation import DEFAULT_TRANSLATION_PROVIDER, TranslationProvider

from typing import Dict, Any, cast, Optional, List
import pandas as pd
from functools import partial

import os

PACKAGE_DATA_DIR = "package_data"
PACKAGE_DATA_PATH = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), PACKAGE_DATA_DIR
)
ALL_REGIONS: pd.DataFrame = pd.read_csv(
    os.path.join(PACKAGE_DATA_PATH, "regions.csv"), index_col="region_id"
)


class ConfigMapping:
    """[summary]

    :param mapping: [description]
    :type mapping: Dict[str, Any]
    """

    def __init__(self, mapping: Dict[str, Any]):

        self._mapping = mapping

    def __getattr__(self, k: str) -> Any:
        return self._mapping[k]

    def __dir__(self):
        return list(self._mapping.keys())

    def __repr__(self):
        return "\n".join(
            name.ljust(30, ".") + " " + id for name, id in self._mapping.items()
        )

    def __iter__(self):
        return self._mapping.values().__iter__()


def hirachy_up(
    lowestids: str, hirachy_frame: pd.DataFrame = ALL_REGIONS
) -> pd.DataFrame:
    """[summary]

    :param lowestids: [description]
    :type lowestids: str
    :param hirachy_frame: [description], defaults to ALL_REGIONS
    :type hirachy_frame: pd.DataFrame, optional
    :raises RuntimeError: [description]
    :raises RuntimeError: [description]
    :return: [description]
    :rtype: pd.DataFrame
    """
    anscestors = []
    current_ids = lowestids
    while len(current_ids) > 0:
        current_regions = hirachy_frame.query("index.isin(@current_ids)")
        anscestors.append(current_regions)
        current_ids = current_regions.dropna().parent.unique()
    return pd.concat(anscestors).sort_index()


def hirachy_down(
    highest_ids: str,
    lowest_level: str = "lau",
    hirachy_frame: pd.DataFrame = ALL_REGIONS,
) -> pd.DataFrame:
    """[summary]

    :param highest_ids: [description]
    :type highest_ids: str
    :param lowest_level: [description], defaults to "lau"
    :type lowest_level: str, optional
    :param hirachy_frame: [description], defaults to ALL_REGIONS
    :type hirachy_frame: pd.DataFrame, optional
    :raises RuntimeError: [description]
    :raises RuntimeError: [description]
    :return: [description]
    :rtype: pd.DataFrame
    """
    descendents = [hirachy_frame.query("index.isin(@highest_ids)")]
    current_ids = highest_ids
    while len(current_ids) > 0:
        current_regions = hirachy_frame.query("parent.isin(@current_ids)")
        descendents.append(current_regions)
        current_ids = current_regions.dropna().index.unique()
        if lowest_level in current_regions.level.unique():
            break
    return pd.concat(descendents).sort_index()


def siblings(
    region_id: pd.DataFrame, hirachy_frame: pd.DataFrame = ALL_REGIONS
) -> pd.DataFrame:
    """[summary]

    :param region_id: [description]
    :type region_id: pd.DataFrame
    :param hirachy_frame: [description], defaults to ALL_REGIONS
    :type hirachy_frame: pd.DataFrame, optional
    :raises RuntimeError: [description]
    :raises RuntimeError: [description]
    :return: [description]
    :rtype: pd.DataFrame
    """
    parent = (  # noqa: F841
        hirachy_frame.query("index == @region_id").loc[:, "parent"].iloc[0]
    )
    return hirachy_frame.query("parent == @parent")


def get_regions() -> pd.DataFrame:
    """List of all the regions and their hierachy structure.

    This function returns a DataFrame of all the regions.
    It contains the name of the region and the its id.
    The latter is required to build queries. Additionally
    information is provided regarding the hierachy structure by
    listing the parent region for each region. Furthermore
    the regions statistical classification (nuts/lau) is provided.
    To allow for more filter options.

    For performance reasons this is simply read from disk.
    The regions are not expected to change significantly over time.
    Nonetheless an up to date DataFrame can be obtained with
    download_all_regions

    :return: DataFrame with all regions.
    """
    return ALL_REGIONS.copy()


state_regions: pd.DataFrame = get_regions().query('level == "nuts1"')
federal_state_dictionary = {
    region.name.replace("-", "_"): region.Index for region in state_regions.itertuples()
}

federal_states = ConfigMapping(federal_state_dictionary)


def get_statistics(
    search: Optional[str] = None,
    stat_meta_data_provider=None,
    target_language: str = "de",
    translation_provider: TranslationProvider = None,
) -> pd.DataFrame:
    """List of all the currently available statistics.

    This frunction returns a DataFrame of all available statistics.
    It contains the statistic code, which is required by the queries.
    It also contains a short and a long description of each statistic.
    By default it returns all available statistics, but it also
    has to option to provide a search keyword in advance.

    The original statistic description are in Germna, but the function
    also allows to get a machine translated version for english of these
    descritpions.

    :param search: Search term used for non-case-sensitive
        search in the long description
    :param translation_provider: Object used for translating the statistics.
        Defaults to  default translation provider if None
    :param target_language: language to translate statistic descriptions to,
        Possible values are currently 'de', 'en' for the default translation
        provider.
    :param stat_meta_data_provider: Source object used to obtain the
        statistic descriptions. Uses global default if missing.
    :return: Table with available statistics.
    """
    if stat_meta_data_provider is None:
        stat_meta_data_provider = DEFAULT_STATISTICS_META_DATA_PROVIDER

    if translation_provider is None:
        translation_provider = DEFAULT_TRANSLATION_PROVIDER

    if target_language != "de" and not translation_provider.is_valid_language_code(
        target_language
    ):
        valid_language_codes = str(translation_provider.get_valid_language_codes())
        raise ValueError(
            "Target language {0} is invalid or not available for translation provider, "
            "please use one of {1}".format(target_language, valid_language_codes)
        )

    stat_descr = stat_meta_data_provider.get_stat_descriptions()

    stat_frame = pd.DataFrame(
        [(stat, *stat_descr[stat]) for stat in stat_descr],
        columns=["statistic", "short_description", "long_description"],
    ).set_index("statistic")

    if target_language != "de":
        translation_provider.translate_data_frame_from_german(
            stat_frame, target_language
        )

    if search is not None:
        search_string = cast(str, search)  # noqa: F841
        return stat_frame.query(
            "short_description.str.contains(@search_string,case=False)"
        )
    else:
        return stat_frame


def get_availability_summary() -> pd.DataFrame:
    """Summary of available data for region/statistic combinations.

    There are many regions and statistics available within the
    datenguide API/at the original sources. Nonetheless data is not
    available for all combinations of statistics and regions.
    Furthermore some statistics might have been discontinued after
    a certain point in time.

    To help with the search for available statistics the function
    proved results from and availablility analysis for all
    statistics and all regions for nuts1, nuts2 and nuts3.
    This function returns the results of this analysis and contains
    for each analyzed region/statistic pair the corresponding
    id/code, the number of entries in the database and if applicable
    the first and last year when this statistic appeared.

    The function does not contain an overview of the lau regions
    and it does not contain an overview of possible drilldowns
    in statstics. For instance is the statstic available for men
    and women individually on top of its availability for the
    combined population.


    :return: Table with available statistics.
    """

    path = os.path.join(PACKAGE_DATA_PATH, "overview.csv")
    return pd.read_csv(path, converters={"region_id": lambda x: str(x)}).set_index(
        ["region_id", "statistic"]
    )


def download_all_regions() -> pd.DataFrame:
    """Downloads all current regions and their hierarchy structure.

    :raises RuntimeError: [description]
    :raises RuntimeError: [description]
    :return: [description]
    :rtype: pd.DataFrame
    """

    def nuts_query(nuts_level):
        q = Query.all_regions(nuts=nuts_level)
        return q

    def lau_query(lau_level):
        q = Query.all_regions(lau=lau_level)
        return q

    qb_all = Query.all_regions()

    qe = QueryExecutioner()
    print("start")
    all_regions = qe.run_query(qb_all)
    print("all")
    r_nuts1 = qe.run_query(nuts_query(1))
    print("nuts1")
    r_nuts2 = qe.run_query(nuts_query(2))
    print("nuts2")
    r_nuts3 = qe.run_query(nuts_query(3))
    print("nuts3")
    r_lau1 = qe.run_query(lau_query(1))
    print("lau")
    # currently no distinction between different laus
    # on datenguide side
    # r_lau2 = qe.run_query(lau_query(2))

    levels = {
        "nuts1": r_nuts1,
        "nuts2": r_nuts2,
        "nuts3": r_nuts3,
        "lau": r_lau1,
        # 'lau2':r_lau2
    }

    def isAnscestor(region_id, candidate):
        """[summary]

        :param region_id: [description]
        :type region_id: [type]
        :param candidate: [description]
        :type candidate: [type]
        :return: [description]
        :rtype: [type]
        """
        return region_id.startswith(candidate) and candidate != region_id

    def parent(region_id, region_details):
        """[summary]

        :param region_id: [description]
        :type region_id: [type]
        :param region_details: [description]
        :type region_details: [type]
        :return: [description]
        :rtype: [type]
        """
        desc = region_details.assign(
            ansc=lambda df: df.index.map(lambda i: isAnscestor(region_id, i))
        ).query("ansc")
        max_lev = desc.level.max()  # noqa: F841
        parent_frame = desc.query("level == @max_lev")
        if not parent_frame.empty:
            return parent_frame.iloc[0, :].name
        else:
            None

    if all_regions is None:
        raise RuntimeError("Was not able to download all regions")

    for k in levels:
        if levels[k] is None:
            raise RuntimeError(f"Was not able to download {k} regions")

    all_regions_df = pd.concat(
        [
            pd.DataFrame(page["data"]["allRegions"]["regions"])
            for page in cast(List[ExecutionResults], all_regions)[0].query_results
        ]
    ).set_index("id")

    level_df = pd.concat(
        pd.concat(
            [
                pd.DataFrame(page["data"]["allRegions"]["regions"])
                for page in cast(List[ExecutionResults], levels[k])[0].query_results
            ]
        ).assign(level=k)
        for k in levels
    )

    all_rg_parents = all_regions_df.join(
        level_df.set_index("id").loc[:, "level"]
    ).assign(
        parent=lambda df: df.index.map(
            partial(
                parent,
                region_details=all_regions_df.assign(
                    level=lambda df: df.index.map(len)
                ),
            )
        )
    )
    all_rg_parents.loc[all_rg_parents.level == "nuts1", "parent"] = "DG"

    return all_rg_parents



from typing import Dict, Any, cast, Optional, NamedTuple, List, Tuple, Union
from typing_extensions import Protocol
import requests
import re

from datenguidepy.schema_json_meta import get_schema_json, get_json_path

Json_Dict = Dict[str, Any]
Json_List = List[Json_Dict]
Json = Union[Json_Dict, Json_List]

StatMeta = Dict[str, str]
UnitMeta = Dict[str, str]
EnumMeta = Dict[str, Dict[Optional[str], str]]
QueryResultsMeta = Dict[str, Union[StatMeta, EnumMeta, UnitMeta]]


class ExecutionResults(NamedTuple):
    """Results of a query with the results itself and the according meta data.
    """

    query_results: Json_List
    meta_data: QueryResultsMeta

    def contains_undefined_region_result(self):
        query_results_with_empty_region = list(
            filter(
                lambda query_result: query_result["data"]["region"] is None,
                self.query_results,
            )
        )
        return len(query_results_with_empty_region) > 0


class TypeMetaData(NamedTuple):
    """The meta data of a field, which consist of the kind, fields and enum values.
    """

    kind: str
    fields: Optional[Json_Dict]
    enum_values: Optional[Dict[str, str]]


class FieldMetaDict(dict):
    """[description]
    """

    def get_return_type(self) -> str:
        """Returns the return type of the field of the FieldMetaDict.

        :return: The return type of the field.
        :rtype: str
        """
        if self["type"]["kind"] == "LIST":
            return self["type"]["ofType"]["name"]
        else:
            return self["type"]["name"]

    def get_arguments(self) -> Dict[str, Tuple[Optional[str], ...]]:
        """[summary]

        :return: [description]
        :rtype: Dict[str, Tuple[Optional[str], ...]]
        """

        def get_type_of(
            argument: Dict[str, Any]
        ) -> Tuple[Optional[str], Optional[str]]:
            if argument["type"]["ofType"]:
                return (
                    argument["type"]["ofType"]["kind"],
                    argument["type"]["ofType"]["name"],
                )
            else:
                return None, None

        return {
            cast(str, arg["name"]): (
                cast(Optional[str], arg.get("type", {}).get("kind", {})),
                cast(Optional[str], arg.get("type", {}).get("name")),
                *get_type_of(arg),
            )
            for arg in self["args"]
        }


def check_http200_body_error(body_json: Json_Dict) -> None:
    if "errors" in body_json:
        raise RuntimeError(
            "Body contains the following error content\n" + str(body_json)
        )


class GraphQlSchemaMetaDataProvider(object):
    """
        The GraphQlSchema meta data priovider helps to obtain
        meta data about the structure of the Graph QL api. As such
        it helps to privde information as to how structurally correct
        queries are build. It does not directly supply information
        about statistics.
    """

    endpoint: str = "https://api-next.datengui.de/graphql"

    REQUEST_HEADER: Dict[str, str] = {"Content-Type": "application/json"}

    _META_DATA_CACHE: Json_Dict = dict()

    _meta_type_info: str = """
        query TypeInfo($type: String!) {
          __type(name: $type) {
            kind
            enumValues {
              name
              description
            }
            fields {
              name
              type {
                ofType {
                  name
                }
                kind
                name
                description
              }
              description
              args {
                name
                type {
                  kind
                  name
                  ofType {
                    name
                    description
                    kind
                  }
                }
              }
            }
          }
        }
    """

    def __init__(self, endpoint=None):
        if endpoint is not None:
            self.endpoint = endpoint

    def get_type_info(
        self, graph_ql_type: str, verbose=False
    ) -> Optional[TypeMetaData]:
        """Returns a json which at top level is a dict with all the
        fields of the type

        :param graph_ql_type: [description]
        :type graph_ql_type: str
        :param verbose: [description], defaults to False
        :type verbose: bool, optional
        :return: [description]
        :rtype: Optional[TypeMetaData]
        """

        if graph_ql_type in self.__class__._META_DATA_CACHE:
            if verbose:
                print("use cache")
            return self.__class__._META_DATA_CACHE[graph_ql_type]
        variables = {"type": graph_ql_type}
        query_json: Json_Dict = {}
        query_json["query"] = self._meta_type_info
        query_json["variables"] = variables
        if verbose:
            print("query REST API")
        info = self._send_request(query_json)
        if info:
            type_kind = info["data"]["__type"]["kind"]

            if type_kind == "OBJECT":
                field_meta: Optional[Json_Dict] = {
                    f["name"]: FieldMetaDict(f)
                    for f in info["data"]["__type"]["fields"]
                }
            else:
                field_meta = None

            if type_kind == "ENUM":
                enum_vals: Optional[Dict[str, str]] = {
                    value["name"]: value["description"]
                    for value in info["data"]["__type"]["enumValues"]
                }
            else:
                enum_vals = None
            type_meta = TypeMetaData(type_kind, field_meta, enum_vals)
            self.__class__._META_DATA_CACHE[graph_ql_type] = type_meta
            return type_meta
        else:
            return None

    def _send_request(self, query_json: Json_Dict) -> Optional[Json_Dict]:
        resp = requests.post(
            self.endpoint, headers=self.REQUEST_HEADER, json=query_json
        )
        if resp.status_code == 200:
            body_json = resp.json()
            check_http200_body_error(body_json)
            return body_json
        else:
            raise RuntimeError(
                self.endpoint
                + "\n"
                + f"No result, got HTML status code {resp.status_code}"
            )


class StatisticsMetaDataProvider(Protocol):
    def get_query_stat_meta(
        self, query_fields_with_types: List[Tuple[str, str]]
    ) -> StatMeta:
        ...

    def get_query_enum_meta(
        self, query_fields_with_types: List[Tuple[str, str]]
    ) -> EnumMeta:
        ...

    def get_query_unit_meta(
        self, query_fields_with_types: List[Tuple[str, str]]
    ) -> UnitMeta:
        ...

    def get_stat_descriptions(self) -> Dict[str, Tuple[str, str]]:
        ...

    def is_statistic(self, stat_candidate: str) -> bool:
        ...


class StatisticsGraphQlMetaDataProvider(object):
    """
        Statistics meta data providers help to supply informations about details
        pertaining to certain statistics that can be obtained via the API. This
        type of meta information is not API specific and can be obtained from
        different sources.
        This particular data provider uses graphql meta data information to provide
        results.
    """

    def __init__(self, endpoint=None):
        self.schema_meta_data_provider = GraphQlSchemaMetaDataProvider(
            endpoint=endpoint
        )

    def get_query_stat_meta(
        self, query_fields_with_types: List[Tuple[str, str]]
    ) -> StatMeta:
        # Region type contains all the statistics fields
        query_fields = [
            field_with_type[0] for field_with_type in query_fields_with_types
        ]
        stat_descriptions = self.get_stat_descriptions()
        stat_meta = {
            stat: stat_descriptions[stat][0]
            for stat in stat_descriptions
            if stat in query_fields
        }
        return stat_meta

    def get_query_unit_meta(
        self, query_fields_with_types: List[Tuple[str, str]]
    ) -> StatMeta:
        return {
            stat: "StatisticsGraphQlMetaDataProvider does not provide unit information."
            for stat, ty in query_fields_with_types
            if self.is_statistic(stat)
        }

    def get_query_enum_meta(
        self, query_fields_with_types: List[Tuple[str, str]]
    ) -> EnumMeta:
        enum_meta: EnumMeta = {}
        for field, field_type in query_fields_with_types:
            type_info = self.schema_meta_data_provider.get_type_info(field_type)
            if type_info is None:
                enum_meta[field] = {"error": "ENUM META DATA COULD NOT BE LOADED"}
            if cast(TypeMetaData, type_info).kind == "ENUM":
                enum_meta[field] = cast(
                    Dict[Optional[str], str], cast(TypeMetaData, type_info).enum_values
                )
        return enum_meta

    @staticmethod
    def _process_stat_meta_data(type_fields: Json_Dict) -> List[Json_Dict]:
        return [
            type_fields[name]
            for name in type_fields
            if "statistics" in type_fields[name].get_arguments()
        ]

    @staticmethod
    def _extract_main_description(description: str) -> str:
        match = re.match(r"^\s*\*\*([^*]*)\*\*", description)
        if match:
            return match.group(1)
        else:
            return "NO DESCRIPTION FOUND"

    def get_stat_descriptions(self) -> Dict[str, Tuple[str, str]]:
        """[summary]

        :return: [description]
        :rtype: [type]
        """
        stat_meta = self.schema_meta_data_provider.get_type_info("Region")
        if stat_meta:
            stat_descriptions = self._create_stat_desc_dic(
                # casting given "Regions" type
                cast(Json_Dict, cast(TypeMetaData, stat_meta).fields)
            )
            return stat_descriptions
        else:
            raise RuntimeError("Meta data provider was anable to fetch statistics")

    def is_statistic(self, stat_candidate: str) -> bool:
        return stat_candidate in self.get_stat_descriptions()

    @staticmethod
    def _create_stat_desc_dic(raw_response: Json_Dict) -> Dict[str, Tuple[str, str]]:
        return dict(
            (
                field["name"],
                (
                    StatisticsGraphQlMetaDataProvider._extract_main_description(
                        field["description"]
                    ),
                    field["description"],
                ),
            )
            for field in StatisticsGraphQlMetaDataProvider._process_stat_meta_data(
                raw_response
            )
        )


class StatisticsSchemaJsonMetaDataProvider(object):
    """
        Statistics meta data providers help to supply informations about details
        pertaining to certain statistics that can be obtained via the API. This
        type of meta information is not API specific and can be obtained from
        different sources.
        This particular data provider the hard copy of a schema file from the SOAP
        cubes that datenguide extracts fron GENESIS and transfers into their API.
    """

    def __init__(self):
        self._full_data_json = [get_schema_json()]

    @property
    def stat_names(self):
        return [
            k
            for stat in get_json_path(self._full_data_json, ["..", "measures"])
            for k in stat.keys()
        ]

    def get_query_stat_meta(
        self, query_fields_with_types: List[Tuple[str, str]]
    ) -> StatMeta:
        fields = [field for field, _ in query_fields_with_types]
        sd = self.get_stat_descriptions()
        return {stat: sd[stat][0] for stat in sd if stat in fields}

    def get_query_unit_meta(
        self, query_fields_with_types: List[Tuple[str, str]]
    ) -> UnitMeta:

        fields = [field for field, _ in query_fields_with_types]

        units = self.get_stat_units()

        return {stat: units[stat] for stat in units if stat in fields}

    def get_query_enum_meta(
        self, query_fields_with_types: List[Tuple[str, str]]
    ) -> EnumMeta:
        enum_meta: EnumMeta = {}
        for field, _ in query_fields_with_types:
            enum_values = get_json_path(
                self._full_data_json,
                ["..", "measures", "..", "dimensions", field, "value_names"],
            )
            if len(enum_values) > 0:
                gesamt_update = {"GESAMT": "Gesamt"}
                enum_meta[field] = dict(enum_values[0], **gesamt_update)
        return enum_meta

    def is_statistic(self, stat_candidate: str) -> bool:
        return stat_candidate in self.stat_names

    def get_stat_units(self) -> Dict[str, str]:
        def get_unit_info(unit_json):
            return unit_json[0]["measure_name_de"]

        stat_names = get_json_path(
            self._full_data_json, ["..", "measures", "..", "name"]
        )
        units = map(
            get_unit_info,
            get_json_path(self._full_data_json, ["..", "measures", "..", "units"]),
        )
        return dict(zip(stat_names, units))

    def get_stat_descriptions(self) -> Dict[str, Tuple[str, str]]:

        stat_descriptions_short = get_json_path(
            self._full_data_json, ["..", "measures", "..", "title_de"]
        )
        stat_descriptions_long = get_json_path(
            self._full_data_json, ["..", "measures", "..", "definition_de"]
        )
        return {
            name: (short, long)
            for name, short, long in zip(
                self.stat_names, stat_descriptions_short, stat_descriptions_long
            )
        }

    def get_enum_values(self) -> Dict[str, Dict[str, str]]:
        names = get_json_path(
            self._full_data_json, ["..", "measures", "..", "dimensions", "..", "name"]
        )
        values = get_json_path(
            self._full_data_json,
            ["..", "measures", "..", "dimensions", "..", "value_names"],
        )
        gesamt_update = {"GESAMT": "Gesamt"}

        return {name: dict(vs, **gesamt_update) for name, vs in zip(names, values)}


DEFAULT_STATISTICS_META_DATA_PROVIDER = StatisticsSchemaJsonMetaDataProvider()


class QueryExecutioner(object):
    """Queries the Datenguide API for data and meta data.

    :param alternative_endpoint: [description], defaults to None
    :type alternative_endpoint: Optional[str], optional
    :return: [description]
    :rtype: None
    """

    REQUEST_HEADER: Dict[str, str] = {"Content-Type": "application/json"}
    endpoint: str = "https://api-next.datengui.de/graphql"

    def __init__(
        self,
        alternative_endpoint: Optional[str] = None,
        statistics_meta_data_provider=None,
    ) -> None:
        if alternative_endpoint:
            self.endpoint = cast(str, alternative_endpoint)

        self.graph_ql_schema_meta_data_provider = GraphQlSchemaMetaDataProvider(
            self.endpoint
        )

        if statistics_meta_data_provider is None:
            self.stat_meta_data_provider = DEFAULT_STATISTICS_META_DATA_PROVIDER
        else:
            self.stat_meta_data_provider = statistics_meta_data_provider

    def get_type_info(
        self, graph_ql_type: str, verbose=False
    ) -> Optional[TypeMetaData]:
        """Returns a json which at top level is a dict with all the
        fields of the type

        :param graph_ql_type: [description]
        :type graph_ql_type: str
        :param verbose: [description], defaults to False
        :type verbose: bool, optional
        :return: [description]
        :rtype: Optional[TypeMetaData]
        """
        return self.graph_ql_schema_meta_data_provider.get_type_info(
            graph_ql_type, verbose
        )

    @staticmethod
    def _pagination_json(page: int) -> Json_Dict:
        return {"page": page, "itemsPerPage": 1000}

    def run_query(self, query) -> Optional[List[ExecutionResults]]:
        """[summary]

        :param query: [description]
        :type query: [type]
        :return: [description]
        :rtype: Optional[List[ExecutionResults]]
        """
        all_results = [
            self._run_single_query_json(query_json, query._get_fields_with_types())
            for query_json in self._generate_post_json(query)
        ]
        if not any(map(lambda r: r is None, all_results)):
            return [cast(ExecutionResults, r) for r in all_results]
        else:
            return None

    def _run_single_query_json(
        self, query_json: Json_Dict, query_fields_with_types: List[Tuple[str, str]]
    ) -> Optional[ExecutionResults]:
        if "allRegions" in [
            field_with_types[0] for field_with_types in query_fields_with_types
        ]:
            results = []
            page = 0
            while True:
                query_json["variables"] = self._pagination_json(page)
                result_page = self._send_request(query_json)
                if result_page is None:
                    return None
                results.append(result_page)
                if (cast(Json_Dict, result_page)["data"]["allRegions"]["page"] + 1) * (
                    cast(Json_Dict, result_page)["data"]["allRegions"]["itemsPerPage"]
                ) >= cast(Json_Dict, result_page)["data"]["allRegions"]["total"]:
                    break
                else:
                    page += 1
        else:
            single_result = self._send_request(query_json)
            if single_result is None:
                return None
            else:
                results = [single_result]

        if results:
            meta: QueryResultsMeta = dict()
            meta["statistics"] = self.stat_meta_data_provider.get_query_stat_meta(
                query_fields_with_types
            )
            meta["enums"] = self.stat_meta_data_provider.get_query_enum_meta(
                query_fields_with_types
            )
            meta["units"] = self.stat_meta_data_provider.get_query_unit_meta(
                query_fields_with_types
            )
            return ExecutionResults(
                query_results=cast(Json_List, results), meta_data=meta
            )
            print(meta)
        else:
            return None

    @staticmethod
    def _generate_post_json(query) -> List[Dict[str, str]]:
        jsons: List[Dict[str, str]] = []
        for query_string in query.get_graphql_query():
            post_json: Json_Dict = dict()
            post_json["query"] = query_string
            jsons.append(post_json)
        return jsons

    def _send_request(self, query_json: Json_Dict) -> Optional[Json_Dict]:
        resp = requests.post(
            self.endpoint, headers=self.REQUEST_HEADER, json=query_json
        )
        if resp.status_code == 200:
            body_json = resp.json()
            check_http200_body_error(body_json)
            return body_json
        else:
            raise RuntimeError(
                self.endpoint
                + "\n"
                + f"No result, got HTML status code {resp.status_code}"
            )



import os
import json


def get_schema_json():
    curr_dir = os.path.dirname(os.path.abspath(__file__))
    data_dir = os.path.join(curr_dir, "package_data")
    schema_json_path = os.path.join(data_dir, "schema.json")
    with open(schema_json_path, "r") as schema_file:
        full_json = json.load(schema_file)

    return full_json


def get_simple_json_path(json, path_list):
    if json is None or len(path_list) == 0:
        return json
    else:
        return get_simple_json_path(json.get(path_list[0]), path_list[1:])


def get_json_path(json, path_list, default="Missing"):
    if json == [] or len(path_list) == 0:
        return json
    else:
        next_val = path_list[0]
        if next_val == "..":
            return [
                elem
                for sub_j in json
                for v in sub_j.values()
                for elem in get_json_path([v], path_list[1:])
            ]
        elif len(path_list) == 1:
            return [sub_j.get(next_val, default) for sub_j in json]
        else:
            sub_jsons = [sub_j for sub_j in json if sub_j.get(next_val) is not None]
            return [
                elem
                for sub_j in sub_jsons
                for elem in get_json_path([sub_j.get(next_val)], path_list[1:])
            ]


if __name__ == "__main__":
    full_json = get_schema_json()
    print(list(get_simple_json_path(full_json, ["22221", "measures", "RLE001"]).keys()))
    stat_names = get_json_path([full_json], ["..", "measures", "..", "name"])
    stat_descr = get_json_path([full_json], ["..", "measures", "..", "definition_de"])

    print(
        list(
            get_json_path([full_json], ["..", "measures", "..", "dimensions", "NAT"])[
                0
            ].keys()
        )
    )
    print(
        get_json_path(
            [full_json], ["..", "measures", "..", "dimensions", "NAAT", "value_names"]
        )
    )



import pandas as pd
import pytest
import os

from datenguidepy.output_transformer import QueryOutputTransformer
from tests.case_construction import construct_execution_results


@pytest.fixture
def query_result():
    abs_path = os.path.abspath(os.path.dirname(__file__))
    example_path = "examples/transformer_example1.json"
    full_path = os.path.join(abs_path, example_path)
    return construct_execution_results(full_path)


@pytest.fixture
def query_results_with_enum():
    abs_path = os.path.abspath(os.path.dirname(__file__))
    example_path = "examples/transformer_example2.json"
    full_path = os.path.join(abs_path, example_path)
    return construct_execution_results(full_path)


@pytest.fixture
def query_results_with_mult_enum():
    abs_path = os.path.abspath(os.path.dirname(__file__))
    example_path = "examples/multiple_enums.json"
    full_path = os.path.join(abs_path, example_path)
    return construct_execution_results(full_path)


def get_abs_path(fname):
    abs_path = os.path.abspath(os.path.dirname(__file__))
    full_path = os.path.join(abs_path, fname)
    return os.path.normpath(full_path)


@pytest.fixture
def query_result_with_autojoin_and_one_enum():
    full_path = get_abs_path("examples/auto_join_one_sided_enum.json")
    return construct_execution_results(full_path)


@pytest.fixture
def query_results_one_statistic_with_units():
    full_path = get_abs_path("examples/one_statistic_with_units.json")
    return construct_execution_results(full_path)


@pytest.fixture
def query_duplicates_for_states_single_stat():
    full_path = get_abs_path("examples/duplicates_for_states_single_stat.json")
    return construct_execution_results(full_path)


@pytest.fixture
def query_duplicates_for_states_multi_stat():
    full_path = get_abs_path("examples/duplicates_for_states_multi_stat.json")
    return construct_execution_results(full_path)


def test_output_transformer_with_one_statistic_and_units(
    query_results_one_statistic_with_units
):
    """check if units were added correctly"""
    qOutTrans = QueryOutputTransformer(query_results_one_statistic_with_units)
    data_transformed = qOutTrans.transform(add_units=True)

    assert data_transformed.loc[0, "TIE003_unit"] == "Anzahl"
    assert data_transformed.columns[5] == "TIE003_unit"


@pytest.fixture
def query_results_multiple_statistics_with_units():
    full_path = get_abs_path("examples/multiple_statistics_with_units.json")
    return construct_execution_results(full_path)


def test_output_transformer_with_multiple_statistics_and_units(
    query_results_multiple_statistics_with_units
):
    """check if units were added correctly"""
    qOutTrans = QueryOutputTransformer(query_results_multiple_statistics_with_units)
    data_transformed = qOutTrans.transform(add_units=True)

    assert data_transformed.iloc[1, range(4, 15, 2)].to_list() == [
        "Prozent",
        "Prozent",
        "Prozent",
        "Prozent",
        "Prozent",
        "kg",
    ]
    assert data_transformed.columns[range(4, 15, 2)].to_list() == [
        "AI0203_unit",
        "AI0204_unit",
        "AI0205_unit",
        "AI0206_unit",
        "AI0207_unit",
        "AI1902_unit",
    ]


def test_output_transformer_defaults(query_result):
    """ start test of output transformer """
    qOutTrans = QueryOutputTransformer(query_result)

    data_transformed = qOutTrans.transform()

    # test whether transformed output data is a dataframe
    assert type(data_transformed) == pd.DataFrame, "transformed data is not a dataframe"

    assert "id" in data_transformed.columns, "no id colum"
    assert "name" in data_transformed.columns, "no name colum"
    assert "year" in data_transformed.columns, "no year colum"
    assert "BEVMK3" in data_transformed.columns, "statistic values are missing"
    assert (
        "BEVMK3_value" not in data_transformed.columns
    ), "old statistics name still present"

    # columns of outdata should not contain json format
    lenlist = len(data_transformed.columns)
    checklist = ["." in data_transformed.columns[x] for x in range(lenlist)]
    assert not any(checklist), "hierarchy not properly transformed"


def test_output_transformer_format_options(query_result, query_results_with_enum):

    qOutTrans = QueryOutputTransformer(query_result)
    data_transformed = qOutTrans.transform(verbose_statistic_names=True)
    assert (
        # "Von der Scheidung betroffene Kinder (BEVMK3)" in data_transformed.columns
        "BEVMK3 (BEVMK3)"
        in data_transformed.columns
    ), "statistic values are missing"

    enum_values = {
        "AFD",
        "B90_GRUENE",
        "CDU",
        "DIELINKE",
        "FDP",
        "SONSTIGE",
        "SPD",
        "GESAMT",
        None,
    }
    enum_descriptions = {
        "AfD",
        "GRÜNE",
        "CDU/CSU",
        "DIE LINKE",
        "FDP",
        "Sonstige Parteien",
        "SPD",
        "Gesamt",
    }

    qOutTrans = QueryOutputTransformer(query_results_with_enum)
    data_transformed = qOutTrans.transform()
    assert set(data_transformed["PART04"]).issubset(enum_values)

    qOutTrans = QueryOutputTransformer(query_results_with_enum)
    data_transformed = qOutTrans.transform(verbose_enum_values=True)
    assert set(data_transformed["PART04"]).issubset(enum_descriptions)

    qOutTrans = QueryOutputTransformer(query_results_with_enum)
    data_transformed = qOutTrans.transform(
        verbose_enum_values=True, verbose_statistic_names=True
    )
    # assert "Gültige Zweitstimmen (WAHL09)" in data_transformed
    assert "WAHL09 (WAHL09)" in data_transformed


def test_output_transformer_format_options_multi_enum(query_results_with_mult_enum):
    qOutTrans = QueryOutputTransformer(query_results_with_mult_enum)
    data_transformed = qOutTrans.transform(verbose_enum_values=False)
    assert data_transformed["ADVNW2"].iloc[0] == "ADVTN420"
    assert data_transformed["ADVNW1"].iloc[0] is None

    data_transformed = qOutTrans.transform(verbose_enum_values=True)
    print(data_transformed.head())
    assert data_transformed["ADVNW2"].iloc[0] == "Grünanlage"
    assert data_transformed["ADVNW1"].iloc[0] == "Gesamt"


def test_output_transformer_auto_join_enum(query_result_with_autojoin_and_one_enum):
    qOutTrans = QueryOutputTransformer(query_result_with_autojoin_and_one_enum)
    data_transformed = qOutTrans.transform(verbose_enum_values=False)
    assert "BEVSTD_GES" in data_transformed
    assert data_transformed.columns.get_loc("BEVSTD_GES") <= 8
    assert list(data_transformed.BEVSTD_GES.unique()) == [None]

    data_transformed = qOutTrans.transform(verbose_enum_values=True)
    assert "BEVSTD_GES" in data_transformed
    assert list(data_transformed.BEVSTD_GES.unique()) == ["Gesamt"]


def test_dumplicate_removal_single(query_duplicates_for_states_single_stat):
    qOutTrans = QueryOutputTransformer(query_duplicates_for_states_single_stat)
    data_transformed = qOutTrans.transform(remove_duplicates=False)
    assert all(data_transformed.name.value_counts() == 2)

    data_transformed = qOutTrans.transform(remove_duplicates=True)
    assert all(data_transformed.name.value_counts() == 1)


def test_dumplicate_removal_multi(query_duplicates_for_states_multi_stat):
    qOutTrans = QueryOutputTransformer(query_duplicates_for_states_multi_stat)
    data_transformed = qOutTrans.transform(remove_duplicates=False)
    assert all(data_transformed.name.value_counts() == 4)

    data_transformed = qOutTrans.transform(remove_duplicates=True)
    assert all(data_transformed.name.value_counts() == 1)



import pytest
import re
import sys
import io
import pandas as pd
from datenguidepy import Field, Query
from datenguidepy.query_execution import (
    FieldMetaDict,
    TypeMetaData,
    GraphQlSchemaMetaDataProvider,
)


@pytest.fixture
def mock_graphqlschemaprovider(monkeypatch):
    def mock_meta_data(self, return_type):
        type_kind = "OBJECT"
        enum_vals = None
        if return_type == "Region":
            field_meta = {
                "id": FieldMetaDict(
                    {
                        "name": "id",
                        "type": {
                            "ofType": None,
                            "kind": "SCALAR",
                            "name": "String",
                            "description": "The `String` scalar type ...",
                        },
                        "description": "Regionalschlüssel",
                        "args": [],
                    }
                ),
                "name": FieldMetaDict(
                    {
                        "args": [],
                        "description": "Name",
                        "name": "name",
                        "type": {
                            "description": "The `String` scalar...",
                            "kind": "SCALAR",
                            "name": "String",
                            "ofType": None,
                        },
                    }
                ),
                "BEV001": FieldMetaDict(
                    {
                        "name": "BEV001",
                        "type": {
                            "ofType": {"name": "BEV001"},
                            "kind": "LIST",
                            "name": None,
                            "description": None,
                        },
                        "description": "Statistik der Geburten",
                        "args": [
                            {
                                "name": "year",
                                "type": {
                                    "kind": "LIST",
                                    "name": None,
                                    "ofType": {
                                        "name": "Int",
                                        "description": "The `Int` scalar type ...",
                                        "kind": "SCALAR",
                                    },
                                },
                            },
                            {
                                "name": "NAT",
                                "type": {
                                    "kind": "LIST",
                                    "name": None,
                                    "ofType": {
                                        "name": "NAT",
                                        "description": "Nationalität",
                                        "kind": "ENUM",
                                    },
                                },
                            },
                        ],
                    }
                ),
            }
        elif return_type == "BEV001":
            field_meta = {
                "id": FieldMetaDict(
                    {
                        "name": "id",
                        "type": {
                            "ofType": None,
                            "kind": "SCALAR",
                            "name": "String",
                            "description": "The `String` scalar...",
                        },
                        "description": "Interne eindeutige ID",
                        "args": [],
                    }
                ),
                "year": FieldMetaDict(
                    {
                        "args": [...],
                        "description": "Jahr des Stichtages",
                        "name": "year",
                        "type": {
                            "description": "The `Int` scalar...",
                            "kind": "SCALAR",
                            "name": "Int",
                            "ofType": None,
                        },
                    }
                ),
                "value": FieldMetaDict(
                    {
                        "args": [...],
                        "description": "Wert",
                        "name": "value",
                        "type": {
                            "description": "The `Float` scalar ...",
                            "kind": "SCALAR",
                            "name": "Float",
                            "ofType": None,
                        },
                    }
                ),
                "NAT": FieldMetaDict(
                    {
                        "args": [...],
                        "description": "Nationalität",
                        "name": "NAT",
                        "type": {
                            "description": "",
                            "kind": "ENUM",
                            "name": "NAT",
                            "ofType": None,
                        },
                    }
                ),
            }
        elif return_type == "NAT":
            type_kind = "ENUM"
            field_meta = None
            enum_vals = {
                "NAA": "Ausländer(innen)",
                "NAD": "Deutsche",
                "GESAMT": "Gesamt",
            }

        else:
            raise RuntimeError("return type not in Monkeypatch")

        return TypeMetaData(type_kind, field_meta, enum_vals)

    monkeypatch.setattr(GraphQlSchemaMetaDataProvider, "get_type_info", mock_meta_data)


@pytest.fixture
def field_default():
    return Field("WAHL09", args={"year": 2017}, fields=["PART04"], return_type="WAHL09")


@pytest.fixture
def query_default():
    return Query.region(region="09", fields=["BEV001"])


@pytest.fixture
def query_with_enum():
    q = Query.region("09", default_fields=False)
    stat = q.add_field("WAHL09")
    stat.add_field("PART04")
    stat.add_field("year")
    stat.add_field("value")
    return q


@pytest.fixture
def field():
    return Field(
        "WAHL09",
        args={"year": 2017},
        fields=["value", "year", "PART04"],
        default_fields=False,
        return_type="WAHL09",
    )


@pytest.fixture
def query():
    return Query.region(region="09", fields=["BEV001"], default_fields=False)


@pytest.fixture
def all_regions_query(field):
    return Query.all_regions(
        parent="11", fields=["id", "name", field], default_fields=False
    )


@pytest.fixture
def complex_query(field):
    return Query.region(region="09", fields=["id", "name", field], default_fields=False)


@pytest.fixture
def more_complex_query(complex_query):
    query = complex_query
    source = Field("source", fields=["title_de"], return_type="Source")
    statistic2 = Field(
        name="BEV001",
        args={"statistics": "R12612"},
        fields=["value", "year", source],
        default_fields=False,
    )
    query.add_field(statistic2)
    return query


def test_get_all_stats_info():
    info = Query.all_regions().get_info()
    assert "name" in info.fields


def test_get_field_info():
    info = Query.all_regions().get_info("BEV001")
    assert "BEVM01" in info.fields


def test_process_query(query_default):
    df = query_default.results()
    assert isinstance(df, pd.DataFrame)


def test_arguments_info_mocked(mock_graphqlschemaprovider, query):
    stat = query.add_field("BEV001")
    info = stat.arguments_info()
    expected_info = """\x1b[4myear\x1b[0m: LIST of type SCALAR(Int)

    \x1b[4mNAT\x1b[0m: LIST of type ENUM(NAT)
    enum values:
    NAA: Ausländer(innen)
    NAD: Deutsche
    GESAMT: Gesamt"""

    assert re.sub(r"\s+", "", info) == re.sub(r"\s+", "", expected_info)


def test_field_info_mocked(mock_graphqlschemaprovider, query):
    stat = query.add_field("BEV001")
    info = stat.fields_info()
    expected_info = """
    id: Interne eindeutige ID
    year: Jahr des Stichtages
    value: Wert
    NAT: Nationalität
    """
    assert re.sub(r"\s+", "", info) == re.sub(r"\s+", "", expected_info)


def test_description_mocked(mock_graphqlschemaprovider, query):
    stat = query.add_field("BEV001")
    descr = stat.description()
    assert descr == "Lebend Geborene"


def test_enum_mocked(mock_graphqlschemaprovider, query):
    stat = query.add_field("BEV001").add_field("NAT")
    info = stat.enum_info()
    expected_info = """
    NAA: Ausländer(innen)
    NAD: Deutsche
    GESAMT: Gesamt"""

    assert re.sub(r"\s+", "", info) == re.sub(r"\s+", "", expected_info)


def test_get_info_stat_mocked(mock_graphqlschemaprovider, query):
    stringio = io.StringIO()
    sys.stdout = stringio
    stat = query.add_field("BEV001")
    stat.get_info()
    info = stringio.getvalue()
    expected_info = """\x1b[1mkind:\x1b[0m
        OBJECT

        \x1b[1mdescription:\x1b[0m
        Lebend Geborene

        \x1b[1marguments:\x1b[0m
        \x1b[4myear\x1b[0m: LIST of type SCALAR(Int)

        \x1b[4mNAT\x1b[0m: LIST of type ENUM(NAT)
        enum values:
        NAA: Ausländer(innen)
        NAD: Deutsche
        GESAMT: Gesamt

        \x1b[1mfields:\x1b[0m
        id: Interne eindeutige ID
        year: Jahr des Stichtages
        value: Wert
        NAT: Nationalität

        \x1b[1menum values:\x1b[0m
        None"""

    assert re.sub(r"\s+", "", info) == re.sub(r"\s+", "", expected_info)


def test_get_fields_with_return_type(field, query_with_enum):
    fields_and_types = field._get_fields_with_types()
    expected_fields_and_types = set(
        [
            ("WAHL09", "WAHL09"),
            ("value", "Float"),
            ("year", "Int"),
            ("PART04", "PART04"),
        ]
    )
    assert set(fields_and_types) == expected_fields_and_types

    fields_and_types = query_with_enum._get_fields_with_types()
    expected_fields_and_types = set(
        [
            ("WAHL09", "WAHL09"),
            ("region", "Region"),
            ("value", "Float"),
            ("year", "Int"),
            ("PART04", "PART04"),
        ]
    )
    assert set(fields_and_types) == expected_fields_and_types



# -*- coding: utf-8 -*-
from datenguidepy.query_builder import Query  # noqa: F401
from datenguidepy.query_builder import Field  # noqa: F401
from datenguidepy.query_helper import get_regions  # noqa: F401
from datenguidepy.query_helper import get_statistics  # noqa: F401
from datenguidepy.query_helper import get_availability_summary  # noqa: F401

import os.path

"""Top-level package for Datenguide Python."""

version_fp = os.path.join(os.path.dirname(os.path.abspath(__file__)), "VERSION")
with open(version_fp) as version_file:
    version = version_file.readline().strip()

__author__ = """CorrelAid"""
__email__ = "packages@correlaid.org"
__version__ = version



from datenguidepy.translation import SchemaTranslationProvider


def test_translation_for_schema_input():
    source_text = "Viehbestand"
    expected_translation = "livestock"
    actual_translation = SchemaTranslationProvider().translate_from_german(
        source_text, "en"
    )
    assert actual_translation == expected_translation


def test_valid_language_codes_not_empty():
    assert len(SchemaTranslationProvider().get_valid_language_codes()) != 0


def test_valid_language_code():
    assert SchemaTranslationProvider().is_valid_language_code("en") is True



import numpy as np
import pandas as pd
from typing import Dict, List, Any, Set, Container, cast

from datenguidepy.query_execution import (
    ExecutionResults,
    StatMeta,
    EnumMeta,
    UnitMeta,
    QueryResultsMeta,
)
import copy


class QueryOutputTransformer:
    """Transforms the query results into a DataFrame.

        :param query_response: Accepts the return type of the query executioner
             in case a non None value was return.
             This is a list of ExecutionResults as some
             python querys may internally be converted into
             several GraphQL queries to be executed,
             returnning one result each.
        :type query_response: List[ExecutionResults]
        """

    def __init__(self, query_response: List[ExecutionResults]) -> None:

        self.query_response = query_response

    @staticmethod
    def _convert_results_to_frame(
        executioner_result: List[ExecutionResults], remove_duplicates: bool = False
    ) -> pd.DataFrame:
        """Converst raw query results to a DataFrame.

        This function converst thre return values from
        query_execution functinoality into a pandas DataFrame.

        :param executioner_result: Raw query results including meta data.
        :return: DataFrame with query results.
        """
        result_frames = []
        for single_query_response in executioner_result:
            for page in single_query_response.query_results:
                result_frames.append(
                    QueryOutputTransformer._convert_regions_to_frame(
                        page, single_query_response.meta_data, remove_duplicates
                    )
                )
        return pd.concat(result_frames)

    @staticmethod
    def _convert_regions_to_frame(
        query_page: Dict[str, Any],
        meta_data: QueryResultsMeta,
        remove_duplicates: bool = True,
    ) -> pd.DataFrame:
        """Converts and combines raw results for one or more regions.

        This result converts region output from the API. The
        Graphql API has two distinct enpoints, one called "region"
        returning results for a single region and one called "allRegions"
        which returns results for multiple regions. This function identifies
        the endpoint that was used and then converts the results for the one
        or more regions that it finds. If multiple regions are found,
        their results are concatenated.

        :param query_page: Single page of API query results as a python dict
            representation of a json.
        :meta_data: Query relevant meta data.
        :return: Converted results possible combined across multiple regions.
        """
        if "region" in query_page["data"]:
            return QueryOutputTransformer._convert_single_results_to_frame(
                query_page["data"]["region"], meta_data, remove_duplicates
            )
        elif "allRegions" in query_page["data"]:
            allRegions = []
            for region in query_page["data"]["allRegions"]["regions"]:
                allRegions.append(
                    QueryOutputTransformer._convert_single_results_to_frame(
                        region, meta_data, remove_duplicates
                    )
                )
            return pd.concat(allRegions)
        else:
            raise RuntimeError(
                "Only queries containing" + '"region" or "regions" can be transformed'
            )

    @staticmethod
    def _convert_single_results_to_frame(
        region_json: Dict[str, Any],
        meta: QueryResultsMeta,
        remove_duplicates: bool = False,
    ) -> pd.DataFrame:
        """Converts a region sub directory of raw output to a dataframe.

        This is the main internal method for converting raw API output
        to dataframes as results are composed of regions. This converts
        a single regions with the idea that results across regions
        can be concatenated. This function contains logic for joining
        data for several statistics in case more than one was queries.
        Furthermore the columns are conveniently sorted to put the most
        important information to the left.


        :param region_json: [description]
        :param meta: [description]
        :raises RuntimeError: The raised error is meant to cover the case
            where quert results were obtained but meta data wasn't possibly
            due to connection problems.
        :return: DataFrame with query results for a single region.
        """
        if "error" in meta["statistics"]:
            raise RuntimeError(
                "No statistics meta data present. Try rerunning the query"
            )
        statistic_frames = [
            QueryOutputTransformer._create_statistic_frame(region_json[stat])
            for stat in cast(StatMeta, meta["statistics"]).keys()
        ]
        if remove_duplicates:
            statistic_frames = [frame.drop_duplicates() for frame in statistic_frames]

        joined_results, join_cols = QueryOutputTransformer._join_statistic_results(
            statistic_frames, list(cast(StatMeta, meta["statistics"]).keys())
        )
        column_order = QueryOutputTransformer._determine_column_order(
            joined_results, join_cols
        )
        general_fields = QueryOutputTransformer._get_general_fields(
            region_json, cast(StatMeta, meta["statistics"])
        )
        for field in general_fields:
            joined_results[field] = region_json[field]

        renamed_results = QueryOutputTransformer._rename_statistic_fields(
            joined_results[general_fields + column_order],
            cast(StatMeta, meta["statistics"]),
        )

        return renamed_results

    @staticmethod
    def _get_general_fields(
        region_json: Dict[str, Any], stat_meta: Dict[str, str]
    ) -> List[str]:
        """Extract non statistic specific fields.

        For the purpouse of arranging dataframe columns this
        fuction extracts all dicionary fields that do not
        contain a statistic in their name.

        :param region_json: Dictionary for a specific region.
        :param stat_meta: Dictionary containg query meta data.
        :return: List of fields without statistics.
        """
        return [
            field
            for field in region_json
            if all(stat not in field for stat in stat_meta.keys())
        ]

    @staticmethod
    def _rename_statistic_fields(
        statistic_result: pd.DataFrame, stat_meta: Dict[str, str]
    ) -> pd.DataFrame:
        """Renames fields containing the statistic values.

        By default all statistic related fields are prefixed
        with the statistic name. As such the reported statistic
        itself has a column name STATISTIC_value. As the value
        is the most central column it is renamed into
        the the simple name STATISTIC.

        :param statistic_result: Results of a query.
        :param stat_meta: Meta data related to the query.
        :return: Results with renamed statistic column.
        """
        rename_mapping = {f"{stat}_value": stat for stat in stat_meta}
        return statistic_result.rename(columns=rename_mapping)

    @staticmethod
    def _create_statistic_frame(statistic_sub_json: Dict[str, Any]) -> pd.DataFrame:
        """Converst a json to a dataframe.

        This function converts the dictionary representation of a json
        to a pandas dataframe. Currenly this uses pandas directly.
        But it might be sensible to implement custom functionality as
        this function is the main reason for the pandas 1.0 requirement.

        :param statistic_sub_json: Python dictionary json representation.
        :return: Dataframe conversion of the dictionary.
        """
        return pd.json_normalize(statistic_sub_json, sep="_", max_level=1)

    @staticmethod
    def _determine_join_columns(statistic_results: List[pd.DataFrame]) -> Set[str]:
        """Dertermines join columns.

        When several statistics are queried this functino
        determines the columns over which to join
        multiple statistics data frames. This will typically
        lead to joining over the year column and enums
        that the statistics have in common.
        Currently has hardcoded exclusion criteria
        to never join across columns containing "value"
        and "source". This is not expected to be a severe
        limmitation as such joins are considered corner cases
        and can be achieved by post-join filters should
        the need arise.

        :param statistic_results: Dataframes for individual statistics
        :return: Columns over which to join.
        """
        candidates = {
            column
            for frame in statistic_results
            for column in frame
            if "value" not in column and "source" not in column
        }
        return {
            candidate
            for candidate in candidates
            if all(candidate in frame for frame in statistic_results)
        }

    @staticmethod
    def _prefix_frame_cols(
        frame: pd.DataFrame, prefix: str, exceptions: Container[str]
    ) -> pd.DataFrame:
        """Prefixes dataframe column names.

        This function prefixes dataframe column names with
        a given prefix but allows for exceptions to
        to be specified, i.e. columns that will not be prefixed.

        :param frame: Dataframe to be prefixed.
        :param prefix: Prefix to be used.
        :param exceptions: Columns that will not be prefixed.
        :return: Dataframe with prefixed columns.
        """
        result_frame = frame.copy()
        result_frame.columns = [
            prefix + "_" + col if col not in exceptions else col
            for col in result_frame.columns
        ]
        return result_frame

    @staticmethod
    def _join_statistic_results(
        statistic_results: List[pd.DataFrame], statistic_names: List[str]
    ) -> tuple:
        """Joins dataframes containing different statistics.

        When joining the frames, columns are first prefixed
        with statistic names.

        :param statistic_results: Dataframes with the statistics to be joined.
        :param statistic_names: Names of the statistics expected to be
            in the same order as the list of statistic results.
        :return: Joined frame and the columns over which was joined.
        """
        assert len(statistic_results) == len(statistic_names)

        join_columns = list(
            QueryOutputTransformer._determine_join_columns(statistic_results)
        )
        result = QueryOutputTransformer._prefix_frame_cols(
            statistic_results[0], statistic_names[0], join_columns
        )

        if len(statistic_results) == 1:
            return result, join_columns
        else:
            for statistic, name in zip(statistic_results[1:], statistic_names[1:]):
                result = result.merge(
                    QueryOutputTransformer._prefix_frame_cols(
                        statistic, name, join_columns
                    ),
                    on=join_columns,
                    how="outer",
                )
            return result, join_columns

    @staticmethod
    def _determine_column_order(
        joined_frame: pd.DataFrame, join_columns: Set[str]
    ) -> List[str]:
        """Determines column order for joined dataframe.

        This function determines a rearrangement of the DataFrame's
        column list, grouping all source columns to the right
        and other information particularly the
        statistics values to the left.

        :param joined_frame: DataFrame with columns for all the
            statistics from the executed query
        :type joined_frame: pd.DataFrame
        :param join_columns: The columns that where used for joining
            different statistics
        :type join_columns: Set[str]
        :return: List of ordered columns
        :rtype: List[str]
        """
        join_col_list = list(join_columns)
        value_columns = [col for col in joined_frame if "value" in col]
        source_cols = [col for col in joined_frame if "source" in col]
        remaining_cols = [
            col
            for col in joined_frame
            if col not in join_columns
            and col not in value_columns
            and col not in source_cols
        ]
        return join_col_list + value_columns + remaining_cols + source_cols

    @staticmethod
    def _make_verbose_statistic_names(
        output: pd.DataFrame, meta: QueryResultsMeta
    ) -> pd.DataFrame:
        """Exchanges statistic column names for short descriptions.

        By default statistic columns display the statistic code.
        This function converts the code to the short description,
        while keeping the code afterward. The aim is to make
        the dataframe more readable.

        :param output: Query results results after conversion to a dataframe.
        :param meta: Query meta data.
        :return: Dataframe with converted column names.
        """
        descriptions = cast(StatMeta, meta["statistics"])
        name_changes = {
            statistic: f"{descriptions[statistic]} ({statistic})"
            for statistic in descriptions
        }
        return output.rename(columns=name_changes)

    @staticmethod
    def _make_verbose_enum_values(
        output: pd.DataFrame, meta: QueryResultsMeta
    ) -> pd.DataFrame:
        """Exchanges enum codes for short descriptions.

        By default enum codes are displayed in enum columns.
        This function converts the codes to short descriptions.
        The aim is to make
        the dataframe more readable.

        :param output: Query results results after conversion to a dataframe.
        :param meta: Query meta data.
        :return: Dataframe with converted column names.
        """
        enum_mappings = copy.deepcopy(cast(EnumMeta, meta["enums"]))
        for enum in enum_mappings:
            enum_mappings[enum][None] = "Gesamt"
        mapped_frame = output.copy()
        for col, description_map in enum_mappings.items():
            if col in mapped_frame:
                mapped_frame[col] = mapped_frame[col].map(description_map)
            else:
                col_name = next(c for c in mapped_frame if c.endswith(col))
                mapped_frame[col_name] = mapped_frame[col_name].map(description_map)
        return mapped_frame

    @staticmethod
    def _add_units(output: pd.DataFrame, meta: QueryResultsMeta) -> pd.DataFrame:
        """Add units from meta_data to DataFrame.

        :param output: DataFrame with results
        :dtype output: pandas.DataFrame
        :param meta: Dictionary containing metadata for query.
        :dtype meta: QueryResultsMeta
        :return: Return DataFrame with results
        :dtype: pandas.DataFrame

        :raise NotImplementedError: More than one statistic in Query
        """

        def add_unit(statistic: str, unit: str):
            if not isinstance(unit, str):
                raise NotImplementedError("Unit is not a single string.")
            mask = output.columns.str.contains(statistic)
            position = int(np.argmax(mask))
            output.insert(loc=position + 1, column=f"{statistic}_unit", value=unit)

        # # ToDo: Uncertain if only one unit is possible per Statistic
        for statistic, unit in cast(UnitMeta, meta["units"].items()):
            add_unit(statistic, unit)
        return output

    def transform(
        self,
        verbose_statistic_names: bool = False,
        verbose_enum_values: bool = False,
        add_units: bool = False,
        remove_duplicates: bool = False,
    ) -> pd.DataFrame:
        """Transform the queries results into a Pandas DataFrame.

        This function allows for different flags that make
        the results more readable by using meta information
        about the query. By default the dataframe is not enrichted by meta
        information assuming an experienced user familiar with a particular statistic.
        For data exploration it is recommended to turn on one or more flags.

        :param verbose_statistic_names: Toggles statistic codes to short descriptions.
        :param verbose_enum_values: Toggles enum codes to descriptions if enum columns
            are present.
        :param add_units: Toggles the addition of a unit column for each statistic to
            make it easier to interpret the numbers.
        :param remove_duplicates: Removes duplicates from query results, i.e. if the
            exact same number has been reported for the same statistic, year, region
            etc. from the same source it gets removed. Such duplications are sometimes
            caused on the API side and this is convenience functionality to remove them.
            The removal happens before potentially joining several different statistics.
        :return: Returns a pandas DataFrame of the queries results.
        """
        output = self._convert_results_to_frame(self.query_response, remove_duplicates)
        if verbose_statistic_names:
            output = self._make_verbose_statistic_names(
                output, self.query_response[0].meta_data
            )
        if verbose_enum_values:
            output = self._make_verbose_enum_values(
                output, self.query_response[0].meta_data
            )
        if add_units:
            output = self._add_units(output, self.query_response[0].meta_data)
        return output



from datenguidepy.query_execution import (
    QueryExecutioner,
    FieldMetaDict,
    TypeMetaData,
    StatisticsGraphQlMetaDataProvider,
    GraphQlSchemaMetaDataProvider,
    StatisticsSchemaJsonMetaDataProvider,
)
from datenguidepy.output_transformer import QueryOutputTransformer
from datenguidepy.query_helper import (
    get_statistics,
    get_regions,
    federal_states,
    get_availability_summary,
)
from datenguidepy.translation import SchemaTranslationProvider

import pytest
from unittest.mock import Mock
from collections import namedtuple
import pandas as pd


@pytest.fixture
def sample_queries():
    q1 = Mock()
    q1.get_graphql_query.return_value = [
        """
    {
        region(id:"05911") {
            id
            name
            BEVMK3 {
                value
                year
            }
        }
    }
    """
    ]
    q1.get_fields.return_value = ["region", "id", "name", "BEVMK3", "value", "year"]
    q1._get_fields_with_types.return_value = [
        ("region", "Region"),
        ("id", "String"),
        ("name", "String"),
        ("BEVMK3", "BEVMK3"),
        ("value", "Float"),
        ("year", "Int"),
    ]

    mq1 = Mock()
    mq1.get_graphql_query.return_value = """
    {
      __type(name: "Region") {
        name
        fields {
          name
          description
          type {
            name
            kind
          }
        }
      }
    }
    """
    SampleQueries = namedtuple("SampleQueries", "data_query1 meta_query1")
    return SampleQueries(q1, mq1)


@pytest.fixture
def type_request_response():
    request = Mock()
    expected_type_info = TypeMetaData(
        kind="ENUM",
        fields=None,
        enum_values={"R12631": "Statistik rechtskräftiger Urteile in Ehesachen"},
    )
    request.return_value = {
        "data": {
            "__type": {
                "kind": "ENUM",
                "name": "BEVMK3Statistics",
                "enumValues": [
                    {
                        "name": "R12631",
                        "description": "Statistik rechtskräftiger Urteile in Ehesachen",
                    }
                ],
                "fields": None,
            }
        }
    }
    return request, expected_type_info


@pytest.fixture
def sample_stat_meta_response():
    return {
        "id": FieldMetaDict(
            {"name": "id", "description": "Regionalschlüssel", "args": []}
        ),
        "name": FieldMetaDict({"name": "name", "description": "Name", "args": []}),
        "AENW01": FieldMetaDict(
            {
                "name": "AENW01",
                "description": "**Description 1** long additional description",
                "args": [
                    {
                        "name": "year",
                        "defaultValue": None,
                        "type": {
                            "kind": "LIST",
                            "name": None,
                            "ofType": {
                                "name": "Int",
                                "description": "LONG DESCRIPTION",
                                "kind": "SCALAR",
                            },
                        },
                    },
                    {
                        "name": "statistics",
                        "defaultValue": None,
                        "type": {
                            "kind": "LIST",
                            "name": None,
                            "ofType": {
                                "name": "AENW01Statistics",
                                "description": "",
                                "kind": "ENUM",
                            },
                        },
                    },
                ],
            }
        ),
        "AENW02": FieldMetaDict(
            {
                "name": "AENW02",
                "description": "Description 2",
                "args": [
                    {
                        "name": "year",
                        "defaultValue": None,
                        "type": {
                            "kind": "LIST",
                            "name": None,
                            "ofType": {
                                "name": "Int",
                                "description": "LONG DESCRIPTION",
                                "kind": "SCALAR",
                            },
                        },
                    },
                    {
                        "name": "statistics",
                        "defaultValue": None,
                        "type": {
                            "kind": "LIST",
                            "name": None,
                            "ofType": {
                                "name": "AENW02Statistics",
                                "description": "",
                                "kind": "ENUM",
                            },
                        },
                    },
                ],
            }
        ),
    }


@pytest.fixture
def stat_description():
    return (
        "**This md bold text should be extracted** everything else should not",
        "This md bold text should be extracted",
    )


def test_filter_stat_metatdata(sample_stat_meta_response):
    processed_stat_meta = StatisticsGraphQlMetaDataProvider._process_stat_meta_data(
        sample_stat_meta_response
    )
    assert all(
        all(("name" in stat, "description" in stat, "args" in stat))
        for stat in processed_stat_meta
    )
    assert all(
        any(arg["name"] == "statistics" for arg in field["args"])
        for field in processed_stat_meta
    )
    assert len(processed_stat_meta) == 2, "too many fields filtered out"


def test_generate_post_json(sample_queries):
    post_jsons = QueryExecutioner._generate_post_json(sample_queries.data_query1)
    sample_queries.data_query1.get_graphql_query.assert_called_once()
    assert "query" in post_jsons[0], "post jsons does not contain a query key"
    assert (
        post_jsons[0]["query"]
        == sample_queries.data_query1.get_graphql_query.return_value[0]
    ), "query not part of the post json"


def test_extract_stat_desc(stat_description):
    extracted_text = StatisticsGraphQlMetaDataProvider._extract_main_description(
        stat_description[0]
    )
    assert (
        extracted_text == stat_description[1]
    ), "extracted text does not match bold part"


def test_create_stat_desc_dic(sample_stat_meta_response):
    desc_dict = StatisticsGraphQlMetaDataProvider._create_stat_desc_dic(
        sample_stat_meta_response
    )
    assert desc_dict["AENW01"][0] == "Description 1", "first dict entry is wrong"
    assert (
        desc_dict["AENW02"][0] == "NO DESCRIPTION FOUND"
    ), "second dict entry is wrong"


def test_get_args(sample_stat_meta_response):
    args = sample_stat_meta_response["AENW01"].get_arguments()
    assert "year" in args, "year argument missing for AENW01"
    assert "statistics" in args, "statistic missing for AEW01"
    assert args["year"] == ("LIST", None, "SCALAR", "Int")
    assert args["statistics"] == ("LIST", None, "ENUM", "AENW01Statistics")


def test_get_type_info_caches_results(type_request_response):
    req_mock, expected_result = type_request_response
    mdp = GraphQlSchemaMetaDataProvider()
    mdp._send_request = req_mock
    mdp.__class__._META_DATA_CACHE = dict()  # clear cache
    assert (
        "BEVMK3Statistics" not in mdp.__class__._META_DATA_CACHE
    ), "statistics should not be in cache"
    res = mdp.get_type_info("BEVMK3Statistics")
    assert res == expected_result, "incorrect response processing"
    mdp._send_request.assert_called_once()
    assert "BEVMK3Statistics" in mdp.__class__._META_DATA_CACHE
    assert (
        mdp.__class__._META_DATA_CACHE["BEVMK3Statistics"] == expected_result
    ), "cache results are wrong"


def test_get_type_info_uses_cached_results(type_request_response):
    req_mock, expected_result = type_request_response
    mdp = GraphQlSchemaMetaDataProvider()
    mdp._send_request = req_mock
    mdp.__class__._META_DATA_CACHE = {"BEVMK3Statistics": expected_result}
    res = mdp.get_type_info("BEVMK3Statistics")
    assert res == expected_result
    mdp._send_request.assert_not_called()


def test_federal_states():
    state_mappings = [
        ("Schleswig_Holstein", "01"),
        ("Hamburg", "02"),
        ("Niedersachsen", "03"),
        ("Bremen", "04"),
        ("Nordrhein_Westfalen", "05"),
        ("Hessen", "06"),
        ("Rheinland_Pfalz", "07"),
        ("Baden_Württemberg, Land", "08"),
        ("Bayern", "09"),
        ("Saarland", "10"),
        ("Berlin", "11"),
        ("Brandenburg", "12"),
        ("Mecklenburg_Vorpommern", "13"),
        ("Sachsen", "14"),
        ("Sachsen_Anhalt", "15"),
        ("Thüringen", "16"),
    ]
    print(federal_states)
    for name, id in state_mappings:
        dic_id = getattr(federal_states, name)
        assert dic_id == id, f"{dic_id},{id}"


def test_regions_overview_table():
    reg = get_regions()
    assert isinstance(reg, pd.DataFrame)
    assert list(reg.columns) == ["name", "level", "parent"]
    assert reg.index.name == "region_id"
    assert reg.shape[0] > 10000


def test_statistic_overview_table():
    stats = get_statistics(
        stat_meta_data_provider=StatisticsSchemaJsonMetaDataProvider()
    )
    assert isinstance(stats, pd.DataFrame)
    assert stats.index.name == "statistic"
    assert list(stats.columns) == ["short_description", "long_description"]
    assert stats.shape[0] > 400


def test_region_statistic_summary_table():
    summary = get_availability_summary()

    assert isinstance(summary, pd.DataFrame)
    assert {"region_id", "statistic", "entries", "start_year", "end_year"}.issubset(
        set(summary.reset_index().columns)
    )


def test_translated_statistic_overview_table():
    untranslated_statistics = get_statistics(
        stat_meta_data_provider=StatisticsSchemaJsonMetaDataProvider()
    )
    translated_statistics = get_statistics(
        stat_meta_data_provider=StatisticsSchemaJsonMetaDataProvider(),
        target_language="en",
        translation_provider=SchemaTranslationProvider(),
    )
    assert untranslated_statistics.shape == translated_statistics.shape
    assert not untranslated_statistics.equals(translated_statistics)


def test_determine_column_order():
    input_columns = ["source_A", "source_B", "stat_A_value", "stat_B_value", "year"]
    input_frame = pd.DataFrame([], columns=input_columns)
    join_columns = set(["year"])

    output = QueryOutputTransformer._determine_column_order(input_frame, join_columns)
    expected_output = ["year", "stat_A_value", "stat_B_value", "source_A", "source_B"]

    assert output == expected_output


def test_prefix_frame_columns():
    cols = ["year", "stat_value", "source"]
    df = pd.DataFrame([], columns=cols)
    output = list(
        QueryOutputTransformer._prefix_frame_cols(
            df, prefix="A", exceptions=["year"]
        ).columns
    )
    expected_output = ["year", "A_stat_value", "A_source"]

    assert output == expected_output


def test_determine_join_columns():
    df_1 = pd.DataFrame(
        [], columns=["year", "source_detail1", "source_detail2", "value"]
    )
    df_2 = pd.DataFrame(
        [], columns=["year", "GES", "source_detail1", "source_detail2", "value"]
    )
    df_3 = pd.DataFrame(
        [], columns=["year", "source_detail1", "source_detail2", "value"]
    )
    frames_inp = [df_1, df_2, df_3]
    output = QueryOutputTransformer._determine_join_columns(frames_inp)
    expected_output = set(["year"])
    assert output == expected_output

    df_1 = pd.DataFrame(
        [], columns=["year", "NAT", "GES", "source_detail1", "source_detail2", "value"]
    )
    df_2 = pd.DataFrame(
        [], columns=["year", "GES", "source_detail1", "source_detail2", "value"]
    )
    frames_inp = [df_1, df_2]
    output = QueryOutputTransformer._determine_join_columns(frames_inp)
    expected_output = set(["year", "GES"])
    assert output == expected_output


def test_get_general_fields():
    meta_dict = {"stat_1": "stat_1 description", "stat_2": "stat_2 description"}
    region_json = {
        "id": "11",
        "stat_1": [],
        "name": "Berlin",
        "stat_2": [{"year": 2000, "value": 1}, {"year": 2001, "value": 2}],
    }
    output = QueryOutputTransformer._get_general_fields(region_json, meta_dict)
    expected_output = ["id", "name"]
    assert output == expected_output


def test_get_query_specific_stat_meta():
    field_type_list = [
        ("id", "String"),
        ("name", "String"),
        ("WAHL09", "WAHL09"),
        ("PART04", "PART04"),
        ("year", "Int"),
        ("value", "Float"),
    ]
    query_stat_meta = StatisticsSchemaJsonMetaDataProvider().get_query_stat_meta(
        field_type_list
    )
    # "Gültige Zweitstimmen" uninformative due to API changes
    expected_stat_meta = {"WAHL09": "Gültige Zweitstimmen"}
    assert query_stat_meta == expected_stat_meta


def test_get_query_specific_enum_meta():
    field_type_list = [
        ("id", "String"),
        ("name", "String"),
        ("WAHL09", "WAHL09"),
        ("PART04", "PART04"),
        ("year", "Int"),
        ("value", "Float"),
    ]
    query_stat_meta = StatisticsSchemaJsonMetaDataProvider().get_query_enum_meta(
        field_type_list
    )
    expected_stat_meta = {
        "PART04": dict(
            [
                ("AFD", "AfD"),
                ("B90_GRUENE", "GRÜNE"),
                ("CDU", "CDU/CSU"),
                ("DIELINKE", "DIE LINKE"),
                ("FDP", "FDP"),
                ("SONSTIGE", "Sonstige Parteien"),
                ("SPD", "SPD"),
                ("GESAMT", "Gesamt"),
            ]
        )
    }
    print(query_stat_meta, expected_stat_meta)
    assert query_stat_meta == expected_stat_meta



import pytest
import re
from datenguidepy import Field, Query
from datenguidepy.query_execution import FieldMetaDict


@pytest.fixture
def patch_return_types(monkeypatch):
    def field_construction_return_types(self, fieldname):
        essential_types = {"year": "Int", "value": "Float", "source": "Source"}
        return essential_types.get(fieldname, "NOT IN MONKEYPATCH")

    monkeypatch.setattr(Field, "_get_return_type", field_construction_return_types)


@pytest.fixture
def mocked_enum_placeholder(monkeypatch):
    def mocked_enum_placeholder(self):
        return "MOCKED ENUM VALUES"

    monkeypatch.setattr(Field, "enum_info", mocked_enum_placeholder)


@pytest.fixture
def mocked_arguments():
    return {
        "WAHL09": FieldMetaDict(
            {
                "name": "WAHL09",
                "type": {
                    "ofType": {"name": "WAHL09"},
                    "kind": "LIST",
                    "name": None,
                    "description": None,
                },
                "description": "Statistik der Geburten",
                "args": [
                    {
                        "name": "year",
                        "type": {
                            "kind": "LIST",
                            "name": None,
                            "ofType": {
                                "name": "Int",
                                "description": "The `Int` scalar type ...",
                                "kind": "SCALAR",
                            },
                        },
                    },
                    {
                        "name": "NAT",
                        "type": {
                            "kind": "LIST",
                            "name": None,
                            "ofType": {
                                "name": "NAT",
                                "description": "Nationalität",
                                "kind": "ENUM",
                            },
                        },
                    },
                ],
            }
        )
    }


@pytest.fixture
def meta_fields():
    field_meta = {
        "id": {
            "name": "id",
            "type": {
                "ofType": None,
                "kind": "SCALAR",
                "name": "String",
                "description": "The `String` scalar...",
            },
            "description": "Interne eindeutige ID",
            "args": [],
        },
        "year": {
            "name": "year",
            "type": {
                "ofType": None,
                "kind": "SCALAR",
                "name": "Int",
                "description": "The `Int`...",
            },
            "description": "Jahr des Stichtages",
            "args": [],
        },
        "value": {
            "name": "value",
            "type": {
                "ofType": None,
                "kind": "SCALAR",
                "name": "Float",
                "description": "The `Float`...",
            },
            "description": "Wert",
            "args": [],
        },
        "source": {
            "name": "source",
            "type": {
                "ofType": None,
                "kind": "OBJECT",
                "name": "Source",
                "description": "",
            },
            "description": "Quellenverweis zur GENESIS Regionaldatenbank",
            "args": [],
        },
        "PART04": {
            "name": "PART04",
            "type": {
                "ofType": None,
                "kind": "ENUM",
                "name": "PART04",
                "description": "Parteien",
            },
            "description": "Parteien",
            "args": [],
        },
    }
    return field_meta


@pytest.fixture
def enum_input():
    return {"NATA": "Ausländer(innen)", "NATD": "Deutsche", "GESAMT": "Gesamt"}


@pytest.fixture
def field_default(patch_return_types):
    return Field("WAHL09", args={"year": 2017}, fields=["PART04"], return_type="WAHL09")


@pytest.fixture
def query_default(patch_return_types):
    return Query.region(region="09", fields=["BEV001"])


@pytest.fixture
def query_with_enum(patch_return_types):
    q = Query.region("09", default_fields=False)
    stat = q.add_field("WAHL09")
    stat.add_field("PART04")
    stat.add_field("year")
    stat.add_field("value")
    return q


@pytest.fixture
def field(patch_return_types):
    return Field(
        "WAHL09",
        args={"year": 2017},
        fields=["value", "year", "PART04"],
        default_fields=False,
        return_type="WAHL09",
    )


@pytest.fixture
def query(patch_return_types):
    return Query.region(region="09", fields=["BEV001"], default_fields=False)


@pytest.fixture
def all_regions_query(field, patch_return_types):
    return Query.all_regions(
        parent="11", fields=["id", "name", field], default_fields=False
    )


@pytest.fixture
def complex_query(field, patch_return_types):
    return Query.region(region="09", fields=["id", "name", field], default_fields=False)


@pytest.fixture
def more_complex_query(complex_query):
    query = complex_query
    source = Field("source", fields=["title_de"], return_type="Source")
    statistic2 = Field(
        name="BEV001",
        args={"statistics": "R12612"},
        fields=["value", "year", source],
        default_fields=False,
    )
    query.add_field(statistic2)
    return query


def test_create_query_is_class_query(query):
    assert isinstance(query, Query)


def test_create_query_class_without_start_filed_raises_error():
    with pytest.raises(TypeError):
        Query()


def test_basic_graphql_string(query):
    graphql_query = query.get_graphql_query()
    assert graphql_query[0] == re.sub(
        " +", " ", """{region (id: "09"){BEV001 }}""".replace("\n", " ")
    )


def test_get_fields_to_query(patch_return_types):
    field = Field(
        name="WAHL09",
        args={"year": 2017},
        fields=["value", "PART04"],
        default_fields=False,
    )
    subfields_string = field._get_fields_to_query(field)
    assert subfields_string == "WAHL09 (year: 2017){value PART04 }"


def test_get_complex_graphql_string(complex_query):
    graphql_query = complex_query.get_graphql_query()
    assert graphql_query[0] == re.sub(
        "    ",
        "",
        """{
            region (id: "09"){
                id name WAHL09 (year: 2017){value year PART04 }
                }
            }""".replace(
            "\n", ""
        ),
    )


def test_get_complex_graphql_string_without_args(patch_return_types):
    field = Field(name="WAHL09", fields=["value"], default_fields=False)
    no_args_query = Query.region(
        region="09", fields=["id", "name", field], default_fields=False
    )

    graphql_query = no_args_query.get_graphql_query()
    assert graphql_query[0] == re.sub(
        "    ",
        "",
        """{
        region (id: "09"){
            id name WAHL09 {value }}
            }""".replace(
            "\n", ""
        ),
    )


def test_get_multiple_stats(more_complex_query):
    graphql_query = more_complex_query.get_graphql_query()
    assert graphql_query[0] == re.sub(
        "    ",
        "",
        """{
        region (id: "09"){
            id name WAHL09 (year: 2017){value year PART04 }
            BEV001 (statistics: R12612){value year source {title_de }}}
        }""".replace(
            "\n", ""
        ),
    )


def test_get_all_fields(query):
    assert query.get_fields() == ["region", "BEV001"]


def test_get_all_fields_complex(more_complex_query):
    assert more_complex_query.get_fields() == [
        "region",
        "id",
        "name",
        "WAHL09",
        "value",
        "year",
        "PART04",
        "BEV001",
        "value",
        "year",
        "source",
        "title_de",
    ]


def test_multiple_filter_args(patch_return_types):
    statistic1 = Field(
        name="BETR09",
        args={"FRUNW2": ["FRUART0111", "FRUART0112"]},
        fields=["FRUNW2", "value", "year"],
        default_fields=False,
    )
    multiple_args_query = Query.region(
        region="02", fields=[statistic1], default_fields=False
    )

    graphql_query = multiple_args_query.get_graphql_query()
    assert graphql_query[0] == re.sub(
        "    ",
        "",
        """{
            region (id: "02"){
                BETR09 (FRUNW2: [FRUART0111, FRUART0112]){FRUNW2 value year }
                }
            }""".replace(
            "\n", ""
        ),
    )


def test_all_regions(all_regions_query):
    graphql_query = all_regions_query.get_graphql_query()[0]
    expected_query = re.sub(
        r"\n\s+",
        "",
        """query ($page : Int, $itemsPerPage : Int) {
            allRegions (page: $page, itemsPerPage: $itemsPerPage){
                regions (parent: "11"){
                    id name WAHL09 (year: 2017){
                        value year PART04 }
                }
                page itemsPerPage total }
        }""",
    )
    assert graphql_query == expected_query


def test_nuts(field):
    query = Query.all_regions(
        parent="11", nuts=3, fields=["id", "name", field], default_fields=False
    )
    graphql_query = query.get_graphql_query()[0]
    assert graphql_query == re.sub(
        r"\n\s+",
        "",
        """query ($page : Int, $itemsPerPage : Int) {
            allRegions (page: $page, itemsPerPage: $itemsPerPage){
                regions (parent: "11", nuts: 3){
                    id name WAHL09 (year: 2017){value year PART04 }
                }
                page itemsPerPage total }
        }""",
    )


def test_lau(field):
    query = Query.all_regions(
        parent="11", lau=3, fields=["id", "name", field], default_fields=False
    )
    graphql_query = query.get_graphql_query()[0]
    assert re.sub(" +", " ", graphql_query.replace("\n", " ")) == re.sub(
        r"\n\s+",
        "",
        """query ($page : Int, $itemsPerPage : Int) {
            allRegions (page: $page, itemsPerPage: $itemsPerPage){
                regions (parent: "11", lau: 3){
                    id name WAHL09 (year: 2017){value year PART04 }
                }
                page itemsPerPage total }
        }""",
    )


def test_filter_for_all(query):
    field = Field(
        name="WAHL09",
        args={"year": 2017, "PART04": "ALL"},
        fields=["value", "PART04"],
        default_fields=False,
        return_type="WAHL09",
    )
    query = Query.region(region="09", fields=["id", "name", field])
    graphql_query = query.get_graphql_query()
    assert graphql_query[0] == re.sub(
        "    ",
        "",
        """{
            region (id: "09"){
                id name WAHL09 (
                    year: 2017, filter:{ PART04: { nin: []}}){
                        value PART04 }
                }
        }""".replace(
            "\n", ""
        ),
    )


def test_add_fields_stepwise(patch_return_types):
    query = Query.region(region="11", default_fields=False)
    statistic1 = query.add_field("BEV001", default_fields=False)
    statistic1.add_field("year")
    statistic2 = Field(
        name="WAHL09",
        args={"year": 2017, "PART04": "ALL"},
        fields=["value", "PART04"],
        default_fields=False,
        return_type="WAHL09",
    )
    query.add_field(statistic2)

    query2 = Query.region(
        region="11",
        fields=[
            Field(
                name="BEV001",
                fields=["year"],
                default_fields=False,
                return_type="BEV001",
            ),
            Field(
                name="WAHL09",
                args={"year": 2017, "PART04": "ALL"},
                fields=["value", "PART04"],
                default_fields=False,
                return_type="WAHL09",
            ),
        ],
        default_fields=False,
    )
    assert query.get_graphql_query() == query2.get_graphql_query()

    graphql_query = query.get_graphql_query()
    assert graphql_query[0] == re.sub(
        "    ",
        "",
        """{
                region (id: "11"){
                    BEV001 {year }
                    WAHL09 (year: 2017, filter:{ PART04: { nin: []}}){value PART04 }
                }
        }""".replace(
            "\n", ""
        ),
    )


def test_add_fields_all_regions(patch_return_types):
    all_reg_query = Query.all_regions(parent="11")
    all_reg_query.add_field("BEV001")

    graphql_query = all_reg_query.get_graphql_query()[0]
    assert graphql_query == re.sub(
        r"\n\s+",
        "",
        """query ($page : Int, $itemsPerPage : Int) {
            allRegions (page: $page, itemsPerPage: $itemsPerPage){
                regions (parent: "11"){
                    id name BEV001 {
                        year value source {title_de valid_from periodicity name url }}
                }
                page itemsPerPage total }
        }""",
    )


def test_add_args_stepwise(patch_return_types):
    query = Query.region(region="11")
    statistic1 = query.add_field("BEV001")
    statistic1.add_field("year")
    statistic1.add_args({"year": 2017})

    query2 = Query.region(
        region="11", fields=[Field(name="BEV001", args={"year": 2017}, fields=["year"])]
    )

    assert query.get_graphql_query() == query2.get_graphql_query()


def test_default_fields(query_default):
    graphql_query = query_default.get_graphql_query()
    assert graphql_query[0] == re.sub(
        "    ",
        "",
        """{region (id: "09"){id name BEV001
            {year value source
            {title_de valid_from periodicity name url }}}}""".replace(
            "\n", " "
        ),
    )


def test_drop_field(query):
    query = query.drop_field("BEV001")
    assert query.get_fields() == ["region"]


def test_drop_field_without_assignment(query):
    query.drop_field("BEV001")
    assert query.get_fields() == ["region"]


def test_drop_field_all_regions(all_regions_query):
    all_regions_query = all_regions_query.drop_field("WAHL09")
    assert all_regions_query.get_fields() == [
        "allRegions",
        "regions",
        "id",
        "name",
        "page",
        "itemsPerPage",
        "total",
    ]


def test_arguments_info_formatting(
    mocked_arguments, mocked_enum_placeholder, field_default
):
    info = field_default._arguments_info_formatter(mocked_arguments)
    expected_info = """\x1b[4myear\x1b[0m: LIST of type SCALAR(Int)

    \x1b[4mNAT\x1b[0m: LIST of type ENUM(NAT)
    enum values:
    MOCKED ENUM VALUES"""

    assert re.sub(r"\s+", "", info) == re.sub(r"\s+", "", expected_info)


def test_fields_info_formatting(meta_fields, field_default):
    info = field_default._fields_info_formatter(meta_fields)
    expected_info = """id: Interne eindeutige ID
    year: Jahr des Stichtages
    value: Wert
    source: Quellenverweis zur GENESIS Regionaldatenbank
    PART04: Parteien"""

    assert re.sub(r"\s+", "", info) == re.sub(r"\s+", "", expected_info)


def test_enum_info_formatting(enum_input, query):
    info = query.add_field("BEV001").add_field("NAT")._enum_info_formatter(enum_input)
    expected_info = """
    NATA: Ausländer(innen)
    NATD: Deutsche
    GESAMT: Gesamt"""

    assert re.sub(r"\s+", "", info) == re.sub(r"\s+", "", expected_info)


def test_invalid_region_name_raises_exception():
    q = Query.region("jodelverein")
    q.add_field("BEV001")
    with pytest.raises(ValueError, match=r"region is invalid."):
        q.results()


def test_missing_statistic_field_raises_exception():
    q = Query.region("09162000")
    with pytest.raises(Exception, match=r"add .* field"):
        q.results()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

with open("datenguidepy/VERSION") as version_file:
    version = version_file.readline().strip()

requirements = ["pandas>=1.0.0", "requests", "typing_extensions"]

setup_requirements = ["pytest-runner"]

test_requirements = ["pytest"]

setup(
    name="datenguidepy",
    version=version,
    packages=find_packages(include=["datenguidepy"]),
    author="CorrelAid",
    author_email="packages@correlaid.org",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Other Audience",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Natural Language :: German",
        "Operating System :: OS Independent",
        "Topic :: Education",
        "Topic :: Sociology",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    description=(
        "Provids easy access to German " + "publically availible regional statistics"
    ),
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="datenguidepy",
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/CorrelAid/datenguide-python",
    zip_safe=False,
)



from datenguidepy.query_execution import StatisticsSchemaJsonMetaDataProvider


def test_get_query_specific_stat_meta():
    field_type_list = [
        ("id", "String"),
        ("name", "String"),
        ("WAHL09", "WAHL09"),
        ("PART04", "PART04"),
        ("year", "Int"),
        ("value", "Float"),
    ]
    query_stat_meta = StatisticsSchemaJsonMetaDataProvider().get_query_stat_meta(
        field_type_list
    )
    expected_stat_meta = {"WAHL09": "Gültige Zweitstimmen"}
    assert query_stat_meta == expected_stat_meta


def test_get_query_specific_enum_meta():
    field_type_list = [
        ("id", "String"),
        ("name", "String"),
        ("WAHL09", "WAHL09"),
        ("PART04", "PART04"),
        ("year", "Int"),
        ("value", "Float"),
    ]
    query_stat_meta = StatisticsSchemaJsonMetaDataProvider().get_query_enum_meta(
        field_type_list
    )
    expected_stat_meta = {
        "PART04": dict(
            [
                ("AFD", "AfD"),
                ("B90_GRUENE", "GRÜNE"),
                ("CDU", "CDU/CSU"),
                ("DIELINKE", "DIE LINKE"),
                ("FDP", "FDP"),
                ("SONSTIGE", "Sonstige Parteien"),
                ("SPD", "SPD"),
                ("GESAMT", "Gesamt"),
            ]
        )
    }
    assert query_stat_meta == expected_stat_meta


def test_get_query_specific_unit_meta():

    field_type_list = [
        ("id", "String"),
        ("name", "String"),
        ("WAHL09", "WAHL09"),
        ("PART04", "PART04"),
        ("year", "Int"),
        ("value", "Float"),
    ]
    query_stat_meta = StatisticsSchemaJsonMetaDataProvider().get_query_unit_meta(
        field_type_list
    )

    expected_stat_meta = {"WAHL09": "Anzahl"}
    assert query_stat_meta == expected_stat_meta
